#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Extension of the IDAES heat exchanger model for transient simulations
"""

__author__ = "Rusty Gentile, John Eslick, Andrew Lee"

from pyomo.environ import Var, Param
from pyomo.dae import DerivativeVar
from pyomo.common.config import ConfigValue
from idaes.core import declare_process_block_class, useDefault
from idaes.core.util.config import DefaultBool
from idaes.core.util.exceptions import ConfigurationError, IdaesError, DynamicError
from .heat_exchanger import HeatExchangerData


@declare_process_block_class(
    "HeatExchangerLumpedCapacitance", doc="0D heat exchanger for transient simulations"
)
class HeatExchangerLumpedCapacitanceData(HeatExchangerData):
    """
    Lumped capacitance HX unit class.
    """

    CONFIG = HeatExchangerData.CONFIG(implicit=True)

    CONFIG.declare(
        "dynamic_heat_balance",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            doc="""Indicates whether heat holdup in the wall material should
be included in the overall energy balance,
**default** - useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - include wall material heat holdup,
**False** - do not include wall material heat holdup.}""",
        ),
    )

    def _add_wall_variables(self):

        # Use the hot side as a reference
        s1_metadata = self.hot_side.config.property_package.get_metadata()

        # Unit system
        temp_units = s1_metadata.get_derived_units("temperature")
        energy_units = s1_metadata.get_derived_units("energy")
        u_units = s1_metadata.get_derived_units("heat_transfer_coefficient")
        area_units = s1_metadata.get_derived_units("area")
        ua_units = u_units * area_units
        r_units = 1 / ua_units

        self.temperature_wall = Var(
            self.flowsheet().config.time,
            initialize=300.0,
            doc="Average wall temperature",
            units=temp_units,
        )
        self.heat_capacity_wall = Param(
            initialize=1.0,
            mutable=True,
            doc="Total heat capacity of heat exchanger material",
            units=energy_units / temp_units,
        )
        self.ua_cold_side = Var(
            self.flowsheet().config.time,
            initialize=1000.0,
            doc="Overall heat transfer coefficient from the cold side",
            units=ua_units,
        )
        self.ua_hot_side = Var(
            self.flowsheet().config.time,
            initialize=1000.0,
            doc="Overall heat transfer coefficient from the hot side",
            units=ua_units,
        )
        self.thermal_resistance_wall = Param(
            initialize=0.0,
            mutable=True,
            doc="Total thermal resistance of heat exchanger material",
            units=r_units,
        )
        self.thermal_fouling_cold_side = Param(
            initialize=0.0,
            mutable=True,
            doc="Total thermal resistance due to fouling on the cold side",
            units=r_units,
        )
        self.thermal_fouling_hot_side = Param(
            initialize=0.0,
            mutable=True,
            doc="Total thermal resistance due to fouling on the hot side",
            units=r_units,
        )
        self.ua_hot_side_to_wall = Var(
            self.flowsheet().config.time,
            initialize=1000.0,
            doc="Overall heat transfer coefficient between the hot side and "
            "the center of the wall",
            units=ua_units,
        )

    def _setup_dynamics(self):
        """
        The Lumped Capacitance Heat Exchanger model is different from other
        unit models in that it allows for part of the heat balance to be dynamic
        whilst the control volumes are steady state. Thus, we need a custom
        _setup_dynamics method.

        Args:
            None

        Returns:
            None
        """
        # Get flowsheet dynamic flag
        dynamic = self.flowsheet().config.dynamic

        # First, check the dynamic_heat_balance argument
        if self.config.dynamic_heat_balance is useDefault:
            # If use default, use flowsheet flag
            self.config.dynamic_heat_balance = dynamic
        elif self.config.dynamic_heat_balance and not dynamic:
            # If True, flowsheet must also be dynamic
            raise DynamicError(
                "{} trying to declare a dynamic model within "
                "a steady-state flowsheet. This is not "
                "supported by the IDAES framework. Try "
                "creating a dynamic flowsheet instead, and "
                "declaring some models as steady-state.".format(self.name)
            )

        # Next, check the dynamic flag
        if self.config.dynamic == useDefault:
            # Use same setting as dynamic_heat_balance
            self.config.dynamic = self.config.dynamic_heat_balance
        elif self.config.dynamic and not self.config.dynamic_heat_balance:
            # If True, must also have dynamic_heat_balance = True
            raise ConfigurationError(
                "{} dynamic can only be True if dynamic_heat_balance "
                "is also True.".format(self.name)
            )

        # Set and validate has_holdup argument
        if self.config.has_holdup == useDefault:
            # Default to same value as dynamic flag
            self.config.has_holdup = self.config.dynamic
        elif self.config.has_holdup is False:
            if self.config.dynamic is True:
                # Dynamic model must have has_holdup = True
                raise ConfigurationError(
                    "{} invalid arguments for dynamic and has_holdup. "
                    "If dynamic = True, has_holdup must also be True "
                    "(was False)".format(self.name)
                )

    def _add_wall_variable_constraints(self):
        @self.Constraint(
            self.flowsheet().config.time,
            doc="Overall heat transfer coefficient for wall temperature equation",
        )
        def ua_hot_side_to_wall_eq(b, t):
            return b.ua_hot_side_to_wall[t] == 1 / (
                1 / b.ua_hot_side[t]
                + b.thermal_fouling_hot_side
                + 0.5 * b.thermal_resistance_wall
            )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Overall heat transfer coefficient equation",
        )
        def ua_total_eq(b, t):
            return b.overall_heat_transfer_coefficient[t] * b.area == 1 / (
                1 / b.ua_hot_side[t]
                + b.thermal_fouling_hot_side
                + b.thermal_resistance_wall
                + b.thermal_fouling_cold_side
                + 1 / b.ua_cold_side[t]
            )

        @self.Constraint(self.flowsheet().config.time, doc="Wall temperature equation")
        def wall_temperature_eq(b, t):
            return (
                b.temperature_wall[t]
                == 0.5
                * (
                    b.hot_side.properties_in[t].temperature
                    + b.hot_side.properties_out[t].temperature
                )
                + b.hot_side.heat[t] / b.ua_hot_side_to_wall[t]
            )

    def activate_dynamic_heat_eq(self):
        """
        Activates the heat holdup term in the overall energy balance for
        transient simulations. Should only be used with dynamic flowsheets
        and if ``dynamic_heat_balance`` is True.

        Args:
            None
        Returns:
            None
        """

        if not self.config.dynamic_heat_balance:
            raise IdaesError(
                "{} heat holdup term cannot be activated "
                "when `dynamic_heat_balance=False`".format(self.name)
            )

        self.unit_heat_balance.deactivate()
        self.dynamic_heat_balance.activate()

    def deactivate_dynamic_heat_eq(self):
        """
        Removes the heat holdup term from the overall energy balance,
        restoring the behavior of the base 0D HeatExchanger model. This may
        be useful for more complex models which need to be solved in several
        stages.

        Args:
            None
        Returns:
            None
        """
        self.dynamic_heat_balance.deactivate()
        self.unit_heat_balance.activate()

    def build(self):
        """
        Building model

        Args:
            None

        Returns:
            None
        """

        super().build()
        self._add_wall_variables()
        self._add_wall_variable_constraints()

        if self.config.dynamic_heat_balance:

            s1_metadata = self.hot_side.config.property_package.get_metadata()
            temp_units = s1_metadata.get_derived_units("temperature")
            time_units = s1_metadata.get_derived_units("time")

            self.dT_wall_dt = DerivativeVar(
                self.temperature_wall,
                wrt=self.flowsheet().config.time,
                doc="Derivative of wall temperature with respect to time",
                units=temp_units / time_units,
            )

            @self.Constraint(
                self.flowsheet().time,
                doc="Unit heat balance equation with the added capacitance term",
            )
            def dynamic_heat_balance(b, t):
                return (
                    b.hot_side.heat[t] + b.cold_side.heat[t]
                    == -b.dT_wall_dt[t] * b.heat_capacity_wall
                )

            self.activate_dynamic_heat_eq()
            self.dT_wall_dt[self.flowsheet().config.time.first()].fix(0)
            self.dT_wall_dt[:].value = 0

    def initialize(self, *args, **kwargs):

        if self.config.dynamic_heat_balance:
            # If the time derivative terms are defined, deactivate them first
            # and reactivate after calling the base initialization method
            self.deactivate_dynamic_heat_eq()
            super().initialize(*args, **kwargs)
            self.activate_dynamic_heat_eq()
        else:
            # Otherwise, just initialize as normal
            super().initialize(*args, **kwargs)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

from collections import OrderedDict

from pandas import DataFrame

from pyomo.environ import value
from pyomo.network import Arc, Port
from pyomo.core.base.var import VarData, Var
from pyomo.core.base.param import Param
from pyomo.core.base.expression import Expression

import idaes.logger as idaeslog
from idaes.core.util.units_of_measurement import report_quantity

_log = idaeslog.getLogger(__name__)

__author__ = "John Eslick, Andrew Lee"


def arcs_to_stream_dict(
    blk, additional=None, descend_into=True, sort=False, prepend=None, s=None
):
    """
    Creates a stream dictionary from the Arcs in a model, using the Arc names as
    keys. This can be used to automate the creation of the streams dictionary
    needed for the ``create_stream_table_dataframe()`` and  ``stream_states_dict()``
    functions.

    Args:
        blk (pyomo.environ._BlockData): Pyomo model to search for Arcs
        additional (dict): Additional states to add to the stream dictionary,
            which aren't represented by arcs in blk, for example feed or
            product streams without Arcs attached or states internal to a unit
            model.
        descend_into (bool): If True, search subblocks for Arcs as well. The
            default is True.
        sort (bool): If True sort keys and return an OrderedDict
        prepend (str): Prepend a string to the arc name joined with a '.'.
            This can be useful to prevent conflicting names when sub blocks
            contain Arcs that have the same names when used in combination
            with descend_into=False.
        s (dict): Add streams to an existing stream dict.

    Returns:
        Dictionary with Arc names as keys and the Arcs as values.

    """
    if s is None:
        s = {}
    for c in blk.component_objects(Arc, descend_into=descend_into):
        key = c.getname()
        if prepend is not None:
            key = ".".join([prepend, key])
        s[key] = c
    if additional is not None:
        s.update(additional)
    if sort:
        s = OrderedDict(sorted(s.items()))
    return s


def stream_states_dict(streams, time_point=0):
    """
    Method to create a dictionary of state block representing stream states.
    This takes a dict with stream name keys and stream values.

    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        time_point : point in the time domain at which to generate stream table
            (default = 0)

    Returns:
        A pandas DataFrame containing the stream table data.
    """
    stream_dict = OrderedDict()

    def _stream_dict_add(sb, n, i=None):
        """add a line to the stream table"""
        if i is None:
            key = n
        else:
            key = "{}[{}]".format(n, i)
        stream_dict[key] = sb

    for n in streams.keys():
        if isinstance(streams[n], Arc):
            for i, a in streams[n].items():
                try:
                    # if getting the StateBlock from the destination port
                    # fails for any reason try the source port. This could
                    # happen if a port does not have an associated
                    # StateBlock. For example a surrogate model may not
                    # use state blocks, unit models may handle physical
                    # properties without state blocks, or the port could
                    # be used to serve the purpose of a translator block.
                    sb = _get_state_from_port(a.ports[1], time_point)
                except:  # pylint: disable=W0702
                    sb = _get_state_from_port(a.ports[0], time_point)
                _stream_dict_add(sb, n, i)
        elif isinstance(streams[n], Port):
            sb = _get_state_from_port(streams[n], time_point)
            _stream_dict_add(sb, n)
        else:
            # _IndexedStateBlock is a private class, so cannot directly test
            # whether  streams[n] is one or not.
            try:
                sb = streams[n][time_point]
            except KeyError as err:
                raise TypeError(
                    f"Either component type of stream argument {streams[n]} "
                    f"is unindexed or {time_point} is not a member of its "
                    f"indexing set."
                ) from err
            _stream_dict_add(sb, n)
    return stream_dict


def create_stream_table_dataframe(
    streams, true_state=False, time_point=0, orient="columns"
):
    """
    Method to create a stream table in the form of a pandas dataframe. Method
    takes a dict with name keys and stream values. Use an OrderedDict to list
    the streams in a specific order, otherwise the dataframe can be sorted
    later.

    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        true_state : indicated whether the stream table should contain the
            display variables define in the StateBlock (False, default) or the
            state variables (True).
        time_point : point in the time domain at which to generate stream table
            (default = 0)
        orient : orientation of stream table. Accepted values are 'columns'
            (default) where streams are displayed as columns, or 'index' where
            stream are displayed as rows.

    Returns:
        A pandas DataFrame containing the stream table data.
    """

    stream_attributes = OrderedDict()
    stream_states = stream_states_dict(streams=streams, time_point=time_point)
    full_keys = []  # List of all rows in dataframe to fill in missing data

    stream_attributes["Units"] = {}

    for key, sb in stream_states.items():
        stream_attributes[key] = {}
        if true_state:
            disp_dict = sb.define_state_vars()
        else:
            disp_dict = sb.define_display_vars()
        for k in disp_dict:
            for row, i in enumerate(disp_dict[k]):
                stream_key = k if i is None else f"{k} {i}"
                quant = report_quantity(disp_dict[k][i])
                stream_attributes[key][stream_key] = quant.m
                if row == 0 or stream_key not in stream_attributes["Units"]:
                    stream_attributes["Units"][stream_key] = quant.u
                if stream_key not in full_keys:
                    full_keys.append(stream_key)

    # Check for missing rows in any stream, and fill with "-" if needed
    for k, v in stream_attributes.items():
        for r in full_keys:
            if r not in v.keys():
                # Missing row, fill with placeholder
                v[r] = "-"

    return DataFrame.from_dict(stream_attributes, orient=orient)


def create_stream_table_ui(
    streams, true_state=False, time_point=0, orient="columns", precision=5
):
    """
    Method to create a stream table in the form of a pandas dataframe. Method
    takes a dict with name keys and stream values. Use an OrderedDict to list
    the streams in a specific order, otherwise the dataframe can be sorted
    later. Note: This function process each stream the same way
    `create_stream_table_dataframe` does.


    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        true_state : indicated whether the stream table should contain the
            display variables define in the StateBlock (False, default) or the
            state variables (True).
        time_point : point in the time domain at which to generate stream table
            (default = 0)
        orient : orientation of stream table. Accepted values are 'columns'
            (default) where streams are displayed as columns, or 'index' where
            stream are displayed as rows.
        precision: rounding the floating numbers to the give precision. Default
            is 5 digits after the floating point.

    Returns:
        A pandas DataFrame containing the stream table data.
    """

    # Variable Types:
    class VariableTypes:
        UNFIXED = "unfixed"
        FIXED = "fixed"
        PARAMETER = "parameter"
        EXPRESSION = "expression"

    stream_attributes = OrderedDict()
    stream_states = stream_states_dict(streams=streams, time_point=time_point)
    full_keys = []  # List of all rows in dataframe to fill in missing data

    stream_attributes["Units"] = {}

    for key, sb in stream_states.items():
        stream_attributes[key] = {}
        if true_state:
            disp_dict = sb.define_state_vars()
        else:
            disp_dict = sb.define_display_vars()
        for k in disp_dict:
            for row, i in enumerate(disp_dict[k]):
                stream_key = k if i is None else f"{k} {i}"

                # Identifying value's variable type
                var_type = None
                if isinstance(disp_dict[k][i], (VarData, Var)):
                    if disp_dict[k][i].fixed:
                        var_type = VariableTypes.FIXED
                    else:
                        var_type = VariableTypes.UNFIXED
                elif isinstance(disp_dict[k][i], Param):
                    var_type = VariableTypes.PARAMETER
                elif isinstance(disp_dict[k][i], Expression):
                    var_type = VariableTypes.EXPRESSION

                quant = report_quantity(disp_dict[k][i])
                stream_attributes[key][stream_key] = (
                    round(quant.m, precision),
                    var_type,
                )
                if row == 0 or stream_key not in stream_attributes["Units"]:
                    stream_attributes["Units"][stream_key] = quant.u

                if stream_key not in full_keys:
                    full_keys.append(stream_key)

    # Check for missing rows in any stream, and fill with "-" if needed
    for k, v in stream_attributes.items():
        for r in full_keys:
            if r not in v.keys():
                # Missing row, fill with placeholder
                v[r] = "-"

    return DataFrame.from_dict(stream_attributes, orient=orient)


def stream_table_dataframe_to_string(stream_table, **kwargs):
    """
    Method to print a stream table from a dataframe. Method takes any argument
    understood by DataFrame.to_string
    """
    # Set some default values for keyword arguments
    na_rep = kwargs.pop("na_rep", "-")
    justify = kwargs.pop("justify", "center")
    # the lambda here could be replaced by "{:#.5g}".format,
    # but arguably that's not as clearly identifiable as a function/callable
    # pylint: disable-next=unnecessary-lambda
    float_format = kwargs.pop("float_format", lambda x: "{:#.5g}".format(x))

    # Print stream table
    return stream_table.to_string(
        na_rep=na_rep, justify=justify, float_format=float_format, **kwargs
    )


def _get_state_from_port(port, time_point):
    """
    Attempt to find a StateBlock-like object connected to a Port. If the
    object is indexed both in space and time, assume that the time index
    comes first.  If no components are assigned to the Port, raise a
    ValueError. If the first component's parent block has no index, raise an
    AttributeError. If different variables on the port appear to be connected
    to different state blocks, raise a RuntimeError.

    Args:
        port (pyomo.network.Port): a port with variables derived from some
            single StateBlock
        time_point : point in the time domain at which to index StateBlock
            (default = 0)

    Returns:
        (StateBlock-like) : an object containing all the components contained
            in the port.
    """
    vlist = list(port.iter_vars())
    states = [v.parent_block().parent_component() for v in vlist]

    if len(vlist) == 0:
        raise ValueError(
            f"No block could be retrieved from Port {port.name} "
            f"because it contains no components."
        )
    # Check the number of indices of the parent property block. If its indexed
    # both in space and time, keep the second, spatial index and throw out the
    # first, temporal index. If that ordering is changed, this method will
    # need to be changed as well.
    try:
        idx = vlist[0].parent_block().index()
    except AttributeError as err:
        raise AttributeError(
            f"No block could be retrieved from Port {port.name} "
            f"because block {vlist[0].parent_block().name} has no index."
        ) from err
    # Assuming the time index is always first and the spatial indices are all
    # the same
    if isinstance(idx, tuple):
        idx = (time_point, vlist[0].parent_block().index()[1:])

    else:
        idx = (time_point,)
    # This method also assumes that ports with different spatial indices won't
    # end up at the same port. Otherwise this check is insufficient.
    if all(states[0] is s for s in states):
        return states[0][idx]
    raise RuntimeError(
        f"No block could be retrieved from Port {port.name} "
        f"because components are derived from multiple blocks."
    )


def generate_table(blocks, attributes, heading=None, exception=True):
    """
    Create a Pandas DataFrame that contains a list of user-defined attributes
    from a set of Blocks.

    Args:
        blocks (dict): A dictionary with name keys and BlockData objects for
            values. Any name can be associated with a block. Use an OrderedDict
            to show the blocks in a specific order, otherwise the dataframe can
            be sorted later.
        attributes (list or tuple of strings): Attributes to report from a
            Block, can be a Var, Param, or Expression. If an attribute doesn't
            exist or doesn't have a valid value, it will be treated as missing
            data.
        heading (list or tuple of strings): A list of strings that will be used
            as column headings. If None the attribute names will be used.
        exception (bool): If True, raise exceptions related to invalid or
            missing indexes. If false missing or bad indexes are ignored and
            None is used for the table value.  Setting this to False allows
            tables where some state blocks have the same attributes with different
            indexing. (default is True)
    Returns:
        (DataFrame): A Pandas dataframe containing a data table
    """
    if heading is None:
        heading = attributes
    st = DataFrame(columns=heading)
    row = [None] * len(attributes)  # not a big deal but save time on realloc
    for key, s in blocks.items():
        for i, a in enumerate(attributes):
            j = None
            if isinstance(a, (list, tuple)):
                # if a is list or tuple, assume index supplied
                try:
                    assert len(a) > 1
                except AssertionError:
                    _log.error(f"An index must be supplided for attribute {a[0]}")
                    raise AssertionError(
                        f"An index must be supplided for attribute {a[0]}"
                    )
                j = a[1:]
                a = a[0]
            v = getattr(s, a, None)
            if j is not None and v is not None:
                try:
                    v = v[j]
                except KeyError:
                    if not exception:
                        v = None
                    else:
                        _log.error(f"{j} is not a valid index of {a}")
                        raise KeyError(f"{j} is not a valid index of {a}")
            try:
                v = value(v, exception=False)
            except TypeError:
                if not exception:
                    v = None
                else:
                    _log.error(f"Cannot calculate value of {a} (may be subscriptable)")
                    raise TypeError(
                        f"Cannot calculate value of {a} (may be subscriptable)"
                    )
            except ZeroDivisionError:
                v = None
            row[i] = v
        st.loc[key] = row
    return st



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Test the basic 0 = z^3 + bz^2 + cz + d calculator."""

import os
import pytest
import pyomo.environ as pyo
import idaes
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available
from idaes.models.properties.modular_properties.eos.ceos_common import cubic_so_path


def cubic_function(z, b, c, d):
    return z**3 + b * z**2 + c * z + d


def derivs(z, b, c, d):
    # Use the tested derivative formulas to make sure the external function
    # keeps returning the right answers. These formulas were confirmed and
    # documented in a jupyter notebook.
    g = [None] * 3
    h = [None] * 6
    g[0] = 1.0 / (-1 + c / z**2 + 2 * d / z**3)
    g[1] = 1.0 / (-2 * z - b + d / z**2)
    g[2] = 1.0 / (-3 * z**2 - 2 * b * z - c)
    """ the hessian structure is:
            b c d
          +------
        b | 0 1 3
        c | - 2 4
        d | - - 5
    """
    h[0] = g[0] ** 3 * (2 * c / z**3 + 6 * d / z**4)
    h[1] = g[0] ** 2 * (2 * c / z**3 * g[1] + 6 * d / z**4 * g[1] - 1 / z**2)
    h[2] = g[1] ** 3 * (2 + 2 * d / z**3)
    h[3] = g[0] ** 2 * (2 * c / z**3 * g[2] + 6 * d / z**4 * g[2] - 2 / z**3)
    h[4] = g[1] ** 2 * (2 * g[2] + 2 * d / z**3 * g[2] - 1 / z**2)
    h[5] = g[2] ** 3 * (6 * z + 2 * b)
    return g, h


@pytest.mark.unit
def test_cubic_eos_available():
    assert cubic_roots_available()


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_general_cubic_root_finder():
    m = pyo.ConcreteModel()
    m.croot_l = pyo.ExternalFunction(library=cubic_so_path, function="cubic_root_l")
    m.croot_h = pyo.ExternalFunction(library=cubic_so_path, function="cubic_root_h")
    param_dict = {
        1: {"b": -3, "c": 0.5, "d": -1, "three": False},
        2: {"b": -3, "c": 0.5, "d": 1, "three": True},
        3: {"b": 1, "c": 0.5, "d": -1, "three": False},
        4: {"b": 1, "c": 0.5, "d": 2, "three": False},
    }

    for k, v in param_dict.items():
        b = v["b"]
        c = v["c"]
        d = v["d"]
        fl, gl, hl = m.croot_l.evaluate_fgh(args=(b, c, d))
        fh, gh, hh = m.croot_h.evaluate_fgh(args=(b, c, d))
        if v["three"]:
            assert fl < fh
        else:
            assert pytest.approx(fl, abs=1e-5) == fh
        assert pytest.approx(0, abs=1e-6) == cubic_function(fl, b, c, d)
        assert pytest.approx(0, abs=1e-6) == cubic_function(fh, b, c, d)
        g, h = derivs(fl, b, c, d)
        for i, ge in enumerate(g):
            assert pytest.approx(ge, abs=1e-5) == gl[i]
        for i, he in enumerate(h):
            assert pytest.approx(he, abs=1e-5) == hl[i]
        g, h = derivs(fh, b, c, d)
        for i, ge in enumerate(g):
            assert pytest.approx(ge, abs=1e-5) == gh[i]
        for i, he in enumerate(h):
            assert pytest.approx(he, abs=1e-5) == hh[i]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module defines the sets of properties used by IDAES.

There is a base set of standard properties which are applicable in most applications,
and it is intended that specialty applications can and will define their own
property sets as required (e.g. electrolyte systems).
"""
from copy import copy

from pyomo.environ import units as pyunits
from pyomo.core.base.units_container import _PyomoUnit

from idaes.core.util.exceptions import PropertyPackageError
import idaes.logger as idaeslog

__author__ = "Dan Gunter <dkgunter@lbl.gov>, Andrew Lee"

_log = idaeslog.getLogger(__name__)


class _PropertyMetadataIndex:
    def __init__(
        self,
        parent,
        idx,
        method: str = None,
        supported: bool = False,
        required: bool = False,
        valid_range=None,
    ):
        self._parent = parent
        self._idx = idx
        self._method = method
        self._supported = supported
        self._required = required
        self._valid_range = None
        self._set_valid_range(
            valid_range
        )  # This method does some basic validation of input
        self._lock_setattr = True
        # TODO: For future, this would be the place to store default scaling information, etc.
        # TODO: Could also define default bounds, nominal values, etc.

    def __setattr__(self, key, value):
        if hasattr(self, "_lock_setattr") and self._lock_setattr is True:
            raise TypeError("Property metadata does not support assignment.")
        super().__setattr__(key, value)

    def __repr__(self):
        return f"{self._parent._doc} ({self._parent._units}%s%s)" % (
            "supported" if self._supported else "",
            "required" if self._required else "",
        )

    @property
    def name(self):
        """
        Name of sub-property as a string
        """
        suffix = ""
        if self._idx != "none":
            suffix = f"_{self._idx}"
        return f"{self._parent.name}{suffix}"

    @property
    def method(self):
        """
        Reference to method that can be called to construct this property and associated
        constraints if using build-on-demand approach.
        """
        return self._method

    @property
    def units(self):
        """
        Units of measurement for this property. This should be a reference to a quantity defined
        in the UnitSet associated with this property package.
        """
        return self._parent.units

    @property
    def supported(self):
        """
        Bool indicating whether this property package supports calculation of this property.
        """
        return self._supported

    @property
    def required(self):
        """
        Bool indicating whether this property package requires calculation of this property
        by another property package.

        This is most commonly used by reaction packages which rely of thermophysical property
        packages to define other properties.
        """
        return self._required

    @property
    def valid_range(self):
        """
        Tuple indicating valid range of values for this property based on the data and method used
        to calculate it. This should take the form of (lower range, upper range) and can be used to
        set and verify bounds on properties before and after solving in order to inform users of any
        cases where the property values are being extrapolated.

        It is strongly recommended that developers set valid ranges for all state variables in
        a property package, as well as any other key properties.
        """
        return self._valid_range

    def set_method(self, meth: str):
        """
        Set method attribute of property.

        Args:
            meth: reference to method required to construct this property

        Returns:
            None
        """
        if meth is None:
            super().__setattr__("_method", None)
        else:
            super().__setattr__("_method", str(meth))

    def set_supported(self, supported: bool = True):
        """
        Set supported attribute of property

        Args:
            supported: bool indicating whether package supports this property

        Returns:
            None
        """
        super().__setattr__("_supported", bool(supported))

    def set_required(self, required: bool = True):
        """
        Set required attribute of property

        Args:
            required: bool indicating whether package requires this property be defined by
            another property package

        Returns:
            None
        """
        super().__setattr__("_required", bool(required))

    def _set_valid_range(self, valid_range: tuple):
        """
        Set valid_range attribute of property. WARNING: users should generally not change
        the valid_range of a property, as this is determined by the data and method used
        to fit the property correlation.

        Args:
            valid_range: 2-tuple with valid range of values for property (lower, upper)

        Returns:
            None
        """
        if valid_range is not None:
            # Do some basic verification
            if not isinstance(valid_range, tuple) or not len(valid_range) == 2:
                raise ValueError(
                    f"valid_range must be a tuple of length 2 (got {valid_range})"
                )
            elif valid_range[0] > valid_range[1]:
                raise ValueError(
                    f"valid_range must be a 2-tuple with form (lower, upper): first value "
                    f"was greater than second value: {valid_range}"
                )

        super().__setattr__("_valid_range", valid_range)

    def update_property(
        self,
        method: str = None,
        required: bool = None,
        supported: bool = None,
        valid_range: tuple = None,
    ):
        """
        Update attributes of this property.

        Args:
            method: method name (str) to be assigned to construct property
            required : bool indicating whether this property package requires this property to be
                defined by another package
            supported : bool indicating whether this property package supports this property
            valid_range: 2-tuple with valid range of values for property (lower, upper)

        Returns:
            None

        Note that if not provided a value, 'supported' is assumed to be True.
        """
        if method is not None:
            self.set_method(method)
        if required is not None:
            self.set_required(required)
        if supported is not None:
            self.set_supported(supported)
        else:
            # Assume supported if not explicitly stated
            # TODO: Reconsider in the future, for now do this for backwards compatibility
            self.set_supported(True)
        if valid_range is not None:
            self._set_valid_range(valid_range)


class PropertyMetadata:
    """
    Metadata object for defining a property.

    This object stores all the metadata associated with a single property, including:

        - name of property
        - documentation of property
        - units of measurement for this property (defined via associated UnitSet)
        - indexing set for sub-properties (e.g. by component or by phase)
        - attributes for sub-properties identifying

            - method that constructs this property and associated constraints (if build-on-demand)
            - whether property is supported by this package
            - whether this package expects this property to be provided by another package

    """

    # TODO: Add optional doc string to metadata objects
    def __init__(
        self,
        name: str = None,
        doc: str = None,
        units: _PyomoUnit = None,
        indices: list = None,
        initialize: dict = None,
    ):
        """
        Construct a PropertyMetadata object.

        Args:
            name: name of property
            doc: doc string for property
            units: units of measurement for property
            indices: list of sub-property indices
            initialize: dict of metadata attributes to set by sub-property


        Supported metadata attributes are:

            - method: str representing a method name to call to construct this property
            - required: bool indicating whether a property package requires this property to be
                        supplied by another property package
            - supported: bool indicating whether a property package supports this property

        Returns:
            None
        """
        if name is None:
            raise TypeError('"name" is required')
        if doc is None:
            doc = name

        self._name = name
        self._doc = doc

        # TODO: Validate units are from UnitSet or dimensionless - needs deprecation
        self._units = units

        # Record indices - needed for building instance specific cases
        self._indices = indices

        self._lock_setattr = True

        # Create entries for indexed sub-properties
        if indices is None or indices is False:
            indices = ["none"]

        for i in indices:
            super().__setattr__("_" + i, _PropertyMetadataIndex(parent=self, idx=i))

        # Initialize values in sub-properties
        if initialize is not None:
            for k, v in initialize.items():
                self[k].update_property(**v)

    def __getitem__(self, key: str):
        if key is None:
            key = "none"
        try:
            return getattr(self, "_" + key)
        except AttributeError:
            raise KeyError(f"Property {self._name} does not have index {key}.")

    def __repr__(self):
        return f"{self._doc} ({self._units})"

    def __setattr__(self, key, value):
        if hasattr(self, "_lock_setattr") and self._lock_setattr is True:
            raise TypeError("Property metadata does not support assignment.")
        super().__setattr__(key, value)

    @property
    def name(self):
        """
        Doc string for property
        """
        return self._name

    @property
    def units(self):
        """
        Units of measurement for this property. This should be a reference to a quantity defined
        in the UnitSet associated with this property package.
        """
        return self._units

    # TODO: An overall update method across multiple indices?


class PropertySetBase:
    """
    Base class for defining property sets.

    This defines the common methods expected of all PropertySets.
    """

    # Define the standard indices for IDAES properties
    _defined_indices = ["comp", "phase", "phase_comp"]

    def __init__(self, parent):
        self._parent_block = parent
        self._defined_properties = []
        self._defined_indices = copy(self.__class__._defined_indices)
        self._lock_setattr = True
        # Find standard properties defined in class and create instance versions
        for i in dir(self.__class__):
            if not i.startswith("_"):
                # Ignore anything starting with "_"
                iobj = getattr(self.__class__, i)
                if isinstance(iobj, PropertyMetadata):
                    units = iobj.units
                    # Check if units is a placeholder string and update if required
                    if isinstance(units, str):
                        units = getattr(self._parent_block.default_units, units)
                    self._add_property(
                        name=i,
                        doc=iobj._doc,
                        units=units,
                        indices=iobj._indices,  # get desired indices from parent
                    )

    def __setattr__(self, key, value):
        if hasattr(self, "_lock_setattr") and self._lock_setattr is True:
            raise TypeError(
                "PropertySets do not support direct assignment. Please use define_property"
            )
        super().__setattr__(key, value)

    def __getitem__(self, key: str):
        n, i = self.get_name_and_index(key)
        try:
            return getattr(self, n)[i]
        except AttributeError:
            raise KeyError(f"Property {key} is not defined in this PropertySet.")

    def __iter__(self):
        for a in self._defined_properties:
            yield getattr(self, a)

    def define_property(
        self,
        name: str,
        doc: str = None,
        units: _PyomoUnit = None,
        indices=None,
        method: str = None,
        required: bool = False,
        supported: bool = True,
        initialize: dict = None,
    ):
        """
        Define a new property called `name`.

        Args:
            name: name of new property (required)
            doc: doc string for property
            units: quantity defined in associated UnitSet defining the units of measurement for property
            indices: list, bool or None. Indicates what sub-property indices should be added. None = use default set,
                     False = unindexed (only 'none' entry)
            method: reference to build-on-demand method for property (optional,
                    only if indices is None or False, default=None)
            supported: bool indicating if package supports this property (optional,
                       only if indices is None or False, default=True)
            required: bool indicating if package requires this property from another package (optional,
                      only if indices is None or False, default=False)
            initialize: dict containing values for method, required and supported by sub-property indices
                        (optional). Cannot combine with method, required or supported arguments.

        Returns:
            None
        """
        if hasattr(self, name):
            raise PropertyPackageError(
                f"A property with the name {name} already exists. Please use update_property "
                "method if you wish to update an existing property's metadata."
            )

        self._add_property(
            name=name,
            doc=doc,
            method=method,
            supported=supported,
            required=required,
            units=units,
            indices=indices,
            initialize=initialize,
        )

    def _add_property(
        self,
        name: str,
        doc: str = None,
        method: str = None,
        required: bool = False,
        supported: bool = False,
        units: _PyomoUnit = None,
        indices: list = None,
        initialize: dict = None,
    ):
        if indices is not None:
            # Cannot assign method, required or supported as we don't know what index they apply to.
            # Raise exception if these are not default values.
            if required or supported or method is not None:
                raise ValueError(
                    "Cannot assign method, required or supported attributes directly to indexed "
                    "properties. Please use the initialize argument instead."
                )
        else:
            # Otherwise, take global arguments and set to initialize "none" index
            if initialize is None:
                initialize = {
                    "none": {
                        "method": method,
                        "required": required,
                        "supported": supported,
                    },
                }
            else:
                if required or supported or method is not None:
                    raise ValueError(
                        "Cannot provide values for initialize and any of method, required or supported "
                        "arguments. Please use the one approach or the other."
                    )

        if indices is None:
            indices = list(self._defined_indices)
            indices.append("none")
        elif indices is False:
            indices = ["none"]

        super().__setattr__(
            name,
            PropertyMetadata(
                name=name,
                doc=doc,
                units=units,
                indices=indices,
                initialize=initialize,
            ),
        )
        self._defined_properties.append(name)

    def check_required_properties(self, other: "PropertySetBase"):
        """
        Check that other property package supports all properties marked as required by this package.

        Args:
            other: PropertySet to check for supported properties

        Returns:
            list of properties required by this package which are not supported by other package
        """
        unsupported = []
        for a in self._defined_properties:
            aobj = getattr(self, a)
            # TODO: Should refactor parent so this is not private
            # pylint: disable-next=protected-access
            for i in aobj._indices:
                if aobj[i].required:
                    try:
                        if not getattr(other, a)[i].supported:
                            unsupported.append(a + "_" + i)
                    except AttributeError:
                        unsupported.append(a + "_" + i)

        return unsupported

    def list_supported_properties(self):
        """
        Return a list of properties supported by this package

        Returns:
            list
        """
        slist = []
        for p in self:
            # TODO: Should refactor parent so this is not private
            # pylint: disable-next=protected-access
            for i in p._indices:
                if p[i].supported:
                    slist.append(p[i])
        return slist

    def list_unsupported_properties(self):
        """
        Return a list of properties not supported by this package

        Returns:
            list
        """
        ulist = []
        for p in self:
            # TODO: Should refactor parent so this is not private
            # pylint: disable-next=protected-access
            for i in p._indices:
                if not p[i].supported:
                    ulist.append(p[i])
        return ulist

    def list_required_properties(self):
        """
        Return a list of properties required by this package

        Returns:
            list
        """
        rlist = []
        for p in self:
            # TODO: Should refactor parent so this is not private
            # pylint: disable-next=protected-access
            for i in p._indices:
                if p[i].required:
                    rlist.append(p[i])
        return rlist

    @property
    def unitset(self):
        """
        Reference to UnitSet associated with this PropertySet (via the parent metadata object).
        """
        return self._parent_block.default_units

    def get_name_and_index(self, property_name: str):
        """
        Separates an indexed property name into the main property and index,

        This method is written assuming the standard indices, and checks for 'phase' and then
        'component' before checking for any custom indices. Developers of custom PropertySets
        may wish to overload this with custom search logic better suited to the set of defined
        indices.

        Returns:
            name, index: strings indicating the name of the base property and indexing set.
        """
        root_name = property_name
        index_name = None
        _defined_indices = {"phase_comp", "phase", "comp"} | set(self._defined_indices)

        _defined_indices = list(reversed(sorted(_defined_indices, key=len)))

        if property_name in self._defined_properties:
            return property_name, None
        else:
            for i in _defined_indices:
                if property_name.endswith("_" + i):
                    nchar = len(i) + 1
                    root_name = property_name[:-nchar]
                    index_name = i
                    break

        if root_name not in self._defined_properties:
            raise ValueError(
                f"Unhandled property: {property_name}. This is mostly likely due to the property not being defined in this PropertySet."
            )

        return root_name, index_name


class StandardPropertySet(PropertySetBase):
    """
    This object defines all the standard properties supported by IDAES, and also allows for
    definition of new properties if required.
    """

    # Concrete definition of all the standard IDAES properties
    # TODO: Should we separate thermophysical and reaction properties?
    # AL: I am inclined to say no - define all of them, and state which are supported
    # This would allow for hybrid packages in the future
    act = PropertyMetadata(
        name="act",
        doc="Chemical Activity",
        units=pyunits.dimensionless,
    )
    act_coeff = PropertyMetadata(
        name="act_coeff",
        doc="Chemical Activity Coefficient",
        units=pyunits.dimensionless,
    )
    compress_fact = PropertyMetadata(
        name="compress_fact",
        doc="Compressiblity Factor",
        units=pyunits.dimensionless,
    )
    compress_fact_crit = PropertyMetadata(
        name="compress_fact_crit",
        doc="Compressiblity Factor at Critical Point",
        units=pyunits.dimensionless,
    )
    conc_mass = PropertyMetadata(
        name="conc_mass",
        doc="Concentration on a Mass Basis",
        units="DENSITY_MASS",
    )
    conc_mol = PropertyMetadata(
        name="conc_mol",
        doc="Concentration on a Molar Basis",
        units="DENSITY_MOLE",
    )
    cp_mass = PropertyMetadata(
        name="cp_mass",
        doc="Constant Pressure Specific Heat Capacity (Mass Basis)",
        units="HEAT_CAPACITY_MASS",
    )
    cp_mol = PropertyMetadata(
        name="cp_mol",
        doc="Constant Pressure Specific Heat Capacity (Molar Basis)",
        units="HEAT_CAPACITY_MOLE",
    )
    cv_mass = PropertyMetadata(
        name="cv_mass",
        doc="Constant Volume Specific Heat Capacity (Mass Basis)",
        units="HEAT_CAPACITY_MASS",
    )
    cv_mol = PropertyMetadata(
        name="cv_mol",
        doc="Constant Volume Specific Heat Capacity (Molar Basis)",
        units="HEAT_CAPACITY_MOLE",
    )
    dens_mass = PropertyMetadata(
        name="dens_mass",
        doc="Density (Mass Basis)",
        units="DENSITY_MASS",
    )
    dens_mass_crit = PropertyMetadata(
        name="dens_mass_crit",
        doc="Mass Density at Critical Point",
        units="DENSITY_MASS",
    )
    dens_mol = PropertyMetadata(
        name="dens_mol",
        doc="Density (Molar Basis)",
        units="DENSITY_MOLE",
    )
    dens_mol_crit = PropertyMetadata(
        name="dens_mol_crit",
        doc="Molar Density at Critical Point",
        units="DENSITY_MOLE",
    )
    diffus = PropertyMetadata(
        name="diffus",
        doc="Diffusivity Coefficient",
        units="DIFFUSIVITY",
    )
    energy_internal_mass = PropertyMetadata(
        name="energy_internal_mass",
        doc="Specific Internal Energy (Mass Basis)",
        units="ENERGY_MASS",
    )
    energy_internal_mol = PropertyMetadata(
        name="energy_internal_mol",
        doc="Specific Internal Energy (Molar Basis)",
        units="ENERGY_MOLE",
    )
    enth_mass = PropertyMetadata(
        name="enth_mass",
        doc="Specific Enthalpy (Mass Basis)",
        units="ENERGY_MASS",
    )
    enth_mol = PropertyMetadata(
        name="enth_mol",
        doc="Specific Enthalpy (Molar Basis)",
        units="ENERGY_MOLE",
    )
    entr_mass = PropertyMetadata(
        name="entr_mass",
        doc="Specific Entropy (Mass Basis)",
        units="ENERGY_MASS",
    )
    entr_mol = PropertyMetadata(
        name="entr_mol",
        doc="Specific Entropy (Molar Basis)",
        units="ENERGY_MOLE",
    )
    flow_mass = PropertyMetadata(
        name="flow_mass",
        doc="Mass Flow Rate",
        units="FLOW_MASS",
    )
    flow_mol = PropertyMetadata(
        name="flow_mol",
        doc="Molar Flow Rate",
        units="FLOW_MOLE",
    )
    flow_vol = PropertyMetadata(
        name="flow_vol",
        doc="Volumetric Flow Rate",
        units="FLOW_VOL",
    )
    fug = PropertyMetadata(
        name="fug",
        doc="Fugacity",
        units="PRESSURE",
    )
    fug_coeff = PropertyMetadata(
        name="fug_coeff",
        doc="Fugacity Coefficient",
        units=pyunits.dimensionless,
    )
    heat_capacity_ratio = PropertyMetadata(
        name="heat_capacity_ratio",
        doc="Heat Capacity Ration",
        units=pyunits.dimensionless,
    )
    gibbs_mass = PropertyMetadata(
        name="gibbs_mass",
        doc="Specific Gibbs Energy (Mass Basis)",
        units="ENERGY_MASS",
    )
    gibbs_mol = PropertyMetadata(
        name="gibbs_mol",
        doc="Specific Gibbs Energy (Molar Basis)",
        units="ENERGY_MOLE",
    )
    isentropic_speed_sound = PropertyMetadata(
        name="isentropic_speed_sound",
        doc="Isentropic Speed of Sound",
        units="VELOCITY",
    )
    isothermal_speed_sound = PropertyMetadata(
        name="isothermal_speed_sound",
        doc="Isothermal Speed of Sound",
        units="VELOCITY",
    )
    henry = PropertyMetadata(
        name="henry",
        doc="Henry Constant",
        units=pyunits.dimensionless,
        # TODO: Units are an issue here, as there are multiple ways to define this
    )
    mass_frac = PropertyMetadata(
        name="mass_frac",
        doc="Mass Fraction",
        units=pyunits.dimensionless,
    )
    mole_frac = PropertyMetadata(
        name="mole_frac",
        doc="Mole Fraction",
        units=pyunits.dimensionless,
    )
    molality = PropertyMetadata(
        name="molality",
        doc="Molality",
        units="MOLALITY",
    )
    mw = PropertyMetadata(
        name="mw",
        doc="Molecular Weight",
        units="MOLECULAR_WEIGHT",
    )
    phase_frac = PropertyMetadata(
        name="phase_frac",
        doc="Phase Fraction",
        units=pyunits.dimensionless,
    )
    prandtl_number = PropertyMetadata(
        name="prandtl_number",
        doc="Prandtl Number",
        units=pyunits.dimensionless,
    )
    pressure = PropertyMetadata(
        name="pressure",
        doc="Pressure",
        units="PRESSURE",
    )
    pressure_bubble = PropertyMetadata(
        name="pressure_bubble",
        doc="Bubble Point Pressure",
        units="PRESSURE",
    )
    pressure_crit = PropertyMetadata(
        name="pressure_crit",
        doc="Pressure at Critical Point",
        units="PRESSURE",
    )
    pressure_dew = PropertyMetadata(
        name="pressure_dew",
        doc="Dew point Pressure",
        units="PRESSURE",
    )
    pressure_osm = PropertyMetadata(
        name="pressure_osm",
        doc="Osmotic Pressure",
        units="PRESSURE",
    )
    pressure_red = PropertyMetadata(
        name="pressure_red",
        doc="Reduced Pressure",
        units=pyunits.dimensionless,
    )
    pressure_sat = PropertyMetadata(
        name="pressure_sat",
        doc="Saturation Pressure",
        units="PRESSURE",
    )
    surf_tens = PropertyMetadata(
        name="surf_tens",
        doc="Surface Tension",
        units="SURFACE_TENSION",
    )
    temperature = PropertyMetadata(
        name="temperature",
        doc="Temperature",
        units="TEMPERATURE",
        indices=False,
    )
    temperature_bubble = PropertyMetadata(
        name="temperature_bubble",
        doc="Bubble Point Temperature",
        units="TEMPERATURE",
        indices=False,
    )
    temperature_crit = PropertyMetadata(
        name="temperature_crit",
        doc="Temperature at Critical Point",
        units="TEMPERATURE",
    )
    temperature_dew = PropertyMetadata(
        name="temperature_dew",
        doc="Dew Point Temperature",
        units="TEMPERATURE",
        indices=False,
    )
    temperature_red = PropertyMetadata(
        name="temperature_red",
        doc="Reduced Temperature",
        units=pyunits.dimensionless,
        indices=False,
    )
    temperature_sat = PropertyMetadata(
        name="temperature_sat",
        doc="Saturation Temperature",
        units="TEMPERATURE",
    )
    therm_cond = PropertyMetadata(
        name="therm_cond",
        doc="Thermal Conductivity",
        units="THERMAL_CONDUCTIVITY",
    )
    visc_d = PropertyMetadata(
        name="visc_d",
        doc="Dynamic Viscosity",
        units="DYNAMIC_VISCOSITY",
    )
    visc_k = PropertyMetadata(
        name="visc_k",
        doc="Kinematic Viscosity",
        units="KINEMATIC_VISCOSITY",
    )
    vol_mol = PropertyMetadata(
        name="vol_mol",
        doc="Molar Volume",
        units="VOLUME_MOLE",
    )
    vol_mol_crit = PropertyMetadata(
        name="vol_mol",
        doc="Molar Volume at Critical Point",
        units="VOLUME_MOLE",
    )
    # Log terms
    log_act = PropertyMetadata(
        name="log_act",
        doc="Log of Activity",
        units=pyunits.dimensionless,
    )
    log_act_coeff = PropertyMetadata(
        name="log_act_coeff",
        doc="Log of Activity Coefficient",
        units=pyunits.dimensionless,
    )
    log_conc_mol = PropertyMetadata(
        name="log_conc_mol",
        doc="Log of Molar Concentration",
        units=pyunits.dimensionless,
    )
    log_mass_frac = PropertyMetadata(
        name="log_mass_frac",
        doc="Log of Mass Fractions",
        units=pyunits.dimensionless,
    )
    log_molality = PropertyMetadata(
        name="log_molality",
        doc="Log of Molality",
        units=pyunits.dimensionless,
    )
    log_mole_frac = PropertyMetadata(
        name="log_mole_frac",
        doc="Log of Mole Fractions",
        units=pyunits.dimensionless,
    )
    log_mole_frac_pbub = PropertyMetadata(
        name="log_mole_frac_pbub",
        doc="Log of Mole Fractions at Bubble Point Pressure",
        units=pyunits.dimensionless,
    )
    log_mole_frac_pdew = PropertyMetadata(
        name="log_mole_frac_pdew",
        doc="Log of Mole Fractions at Dew Point Pressure",
        units=pyunits.dimensionless,
    )
    log_mole_frac_tbub = PropertyMetadata(
        name="log_mole_frac_tbub",
        doc="Log of Mole Fractions at Bubble Point Temperature",
        units=pyunits.dimensionless,
    )
    log_mole_frac_tdew = PropertyMetadata(
        name="log_mole_frac_tdew",
        doc="Log of Mole Fractions at Dew Point Temperature",
        units=pyunits.dimensionless,
    )
    log_pressure = PropertyMetadata(
        name="log_pressure",
        doc="Log of Pressure",
        units=pyunits.dimensionless,
    )

    # Reaction Properties
    # TODO: Units are also problematic here - no single definition
    dh_rxn = PropertyMetadata(
        name="dh_rxn",
        doc="Heat of Reaction",
        units="ENERGY_MOLE",
    )
    k_eq = PropertyMetadata(
        name="k_eq",
        doc="Equilibrium Coefficient",
        units=pyunits.dimensionless,
    )
    log_k_eq = PropertyMetadata(
        name="log_k_eq",
        doc="Log of Equilibrium Coefficient",
        units=pyunits.dimensionless,
    )
    k_rxn = PropertyMetadata(
        name="k_rxn",
        doc="Rate Constant",
        units=pyunits.dimensionless,
    )
    reaction_rate = PropertyMetadata(
        name="reaction_rate",
        doc="Reaction Rate",
        units=pyunits.dimensionless,
    )


class ElectrolytePropertySet(StandardPropertySet):
    """
    This object defines all the standard properties supported by IDAES, and also allows for
    definition of new properties if required.
    """

    _defined_indices = [
        "comp",
        "phase",
        "phase_comp",
        "phase_comp_apparent",
        "phase_comp_true",
    ]
    # Definition of additional properties required for electrolyte applications
    # Log terms
    log_act_phase_solvents = PropertyMetadata(
        name="log_act_phase_solvents",
        doc="Log of Activity of Solvents",
        units=pyunits.dimensionless,
    )
    ionic_strength = PropertyMetadata(
        name="ionic_strength",
        doc="Ionic Strength (Molality Basis)",
        units="MOLALITY",
    )
    pH = PropertyMetadata(
        name="pH",
        doc="pH, -log10(a_H+)",
        units=pyunits.dimensionless,
    )
    pK = PropertyMetadata(
        name="pK",
        doc="pK, -log10(equilibrium constant)",
        units=pyunits.dimensionless,
    )
    pOH = PropertyMetadata(
        name="pOH",
        doc="pOH, -log10(a_OH-)",
        units=pyunits.dimensionless,
    )
    log10_act_coeff = PropertyMetadata(
        name="log10_act_coeff",
        doc="Log of Activity Coefficient",
        units=pyunits.dimensionless,
    )
    log10_molality = PropertyMetadata(
        name="log10_molality",
        doc="Log of Molality",
        units=pyunits.dimensionless,
    )
    log10_k_eq = PropertyMetadata(
        name="log10_k_eq",
        doc="Log of equilibrium coefficient",
        units=pyunits.dimensionless,
    )
    saturation_index = PropertyMetadata(
        name="saturation_index",
        doc="Saturation Index = log(IAP/Ksp)",
        units=pyunits.dimensionless,
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from abc import abstractmethod
from copy import deepcopy
import numpy as np

from ...util.util import myArrayEq
from .lattice import Lattice
from ..transform_func import TransformFunc


class UnitCell(object):
    DBL_TOL = 1e-5

    # === STANDARD CONSTRUCTOR
    def __init__(self, argTiling, FracPositions):
        self._Tiling = argTiling
        self._FracPositions = FracPositions

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        for Position in self.FracPositions:
            for coord in Position:
                if coord < 0.0 - UnitCell.DBL_TOL or coord > 1.0 + UnitCell.DBL_TOL:
                    return False
        return True

    # === MANIPULATION METHODS
    def addPosition(self, P):
        self._FracPositions.append(P)

    def applyTransF(self, TransF):
        if isinstance(TransF, TransformFunc):
            self._Tiling.applyTransF(TransF)
        else:
            raise TypeError

    # === PROPERTY EVALUATION METHODS
    @property
    def Tiling(self):
        return self._Tiling

    @property
    def FracPositions(self):
        return self._FracPositions

    def convertToFrac(self, P, blnDiscardIntPart=True):
        P[:] = self.Tiling.getFractionalCoords(
            P, blnRelativeToCenter=False, blnRoundInside=True, blnPreferZero=True
        )
        if blnDiscardIntPart:
            P -= P.astype(int)

    def getConvertToFrac(self, P, blnDiscardIntPart=True):
        result = deepcopy(P)
        self.convertToFrac(result, blnDiscardIntPart)
        return result

    def getPointType(self, P):
        PFrac = self.getConvertToFrac(P, blnDiscardIntPart=True)
        for i, FracPosition in enumerate(self.FracPositions):
            if myArrayEq(FracPosition, PFrac, UnitCell.DBL_TOL):
                return i
        return None


class UnitCellLattice(Lattice):
    # === AUXILIARY METHODS
    def ScanRef(self, RefScanMin, RefScanMax):
        for n1 in range(RefScanMin[0], RefScanMax[0] + 1):
            for n2 in range(RefScanMin[1], RefScanMax[1] + 1):
                for n3 in range(RefScanMin[2], RefScanMax[2] + 1):
                    for FracPart in self.RefUnitCell.FracPositions:
                        yield (
                            n1 * self.RefUnitCell.Tiling.TileShape.Vx
                            + n2 * self.RefUnitCell.Tiling.TileShape.Vy
                            + n3 * self.RefUnitCell.Tiling.TileShape.Vz
                            + FracPart
                        )

    def Scan(self, argPolyhedron):
        RefScanMin = self._getConvertToReference(argPolyhedron.V[0])
        RefScanMax = self._getConvertToReference(argPolyhedron.V[0])
        for v in argPolyhedron.V:
            RefScanMin = np.minimum(RefScanMin, self._getConvertToReference(v))
            RefScanMax = np.maximum(RefScanMax, self._getConvertToReference(v))
        RefScanMin = RefScanMin.astype(int) + np.array([-1, -1, -1], dtype=int)
        RefScanMax = RefScanMax.astype(int) + np.array([1, 1, 1], dtype=int)
        for RefP in self.ScanRef(RefScanMin, RefScanMax):
            yield self._getConvertFromReference(RefP)

    # === STANDARD CONSTRUCTOR
    def __init__(self, RefUnitCell):
        self._RefUnitCell = RefUnitCell
        Lattice.__init__(self)

    # === PROPERTY EVALUATION METHODS
    def isOnLattice(self, P):
        return (
            self.RefUnitCell.getPointType(Lattice._getConvertToReference(self, P))
            is not None
        )

    @abstractmethod
    def areNeighbors(self, P1, P2):
        raise NotImplementedError

    @abstractmethod
    def getNeighbors(self, P, layer):
        raise NotImplementedError

    # === BASIC QUERY METHODS
    @property
    def RefUnitCell(self):
        return self._RefUnitCell

    @property
    def UnitCell(self):
        result = deepcopy(self.RefUnitCell)
        for f in self._TransformFuncs:
            result.applyTransF(f)
        return result



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Sub-methods for eNRTL activity coefficient method.

Includes temperature dependence rules for alpha and tau
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Reals, units as pyunits, Var

from idaes.core.util.exceptions import BurntToast, ConfigurationError
import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)


class ConstantAlpha(object):
    """Class for methods assuming constant alpha"""

    @staticmethod
    def build_parameters(b):
        param_block = b.parent_block()

        # Get user provided values for alpha (if present)
        try:
            alpha_data = param_block.config.parameter_data[b.local_name + "_alpha"]
        except KeyError:
            alpha_data = {}

        # Check for unused parameters in alpha_data
        for i, j in alpha_data.keys():
            if (i, j) not in b.component_pair_set_symmetric and (
                j,
                i,
            ) not in b.component_pair_set_symmetric:
                raise ConfigurationError(
                    "{} eNRTL alpha parameter provided for invalid "
                    "component pair {}. Please check typing and only provide "
                    "parameters for valid species pairs.".format(b.name, (i, j))
                )

        def alpha_init(b, i, j):
            if (i, j) in alpha_data.keys():
                v = alpha_data[(i, j)]
                # Check for non-symmetric value assignment
                if (j, i) in alpha_data.keys():
                    if alpha_data[(j, i)] != v:
                        raise ConfigurationError(
                            "{} eNRTL alpha parameter assigned non-symmetric "
                            "value for pair {}. Please assign only one value "
                            "for component pair.".format(b.name, (i, j))
                        )
                    else:
                        _log.info(
                            "eNRTL alpha value provided for both {} and "
                            "{}. It is only necessary to provide a "
                            "value for one of these due to symmetry.".format(
                                (i, j), (j, i)
                            )
                        )
            elif (j, i) in alpha_data.keys():
                v = alpha_data[(j, i)]
            elif (i in param_block.solvent_set or i in param_block.solute_set) and (
                j in param_block.solvent_set or j in param_block.solute_set
            ):
                # Molecular-molecular interaction, default value is 0.3
                v = 0.3
            else:
                # All other interactions have default value 0.2
                v = 0.2
            return v

        b.add_component(
            "alpha",
            Var(
                b.component_pair_set_symmetric,
                within=Reals,
                initialize=alpha_init,
                doc="Symmetric non-randomness parameters",
                units=pyunits.dimensionless,
            ),
        )

    @staticmethod
    def return_expression(b, pobj, i, j, T):
        if (i, j) in pobj.alpha:
            return pobj.alpha[i, j]
        elif (j, i) in pobj.alpha:
            return pobj.alpha[j, i]
        elif i == j:
            return 0.2
        else:
            raise BurntToast(
                "{} alpha rule encountered unexpected index {}. Please contact"
                "the IDAES Developers with this bug.".format(b.name, (i, j))
            )


class ConstantTau(object):
    """Class for methods assuming consant tau"""

    @staticmethod
    def build_parameters(b):
        param_block = b.parent_block()

        # Get user provided values for tau (if present)
        try:
            tau_data = param_block.config.parameter_data[b.local_name + "_tau"]
        except KeyError:
            tau_data = {}

        # Check for unused parameters in tau_data
        for i, j in tau_data.keys():
            if (i, j) not in b.component_pair_set:
                raise ConfigurationError(
                    "{} eNRTL tau parameter provided for invalid "
                    "component pair {}. Please check typing and only provide "
                    "parameters for valid species pairs.".format(b.name, (i, j))
                )

        def tau_init(b, i, j):
            if (i, j) in tau_data.keys():
                v = tau_data[(i, j)]
            else:
                # Default interaction value is 0
                v = 0
            return v

        b.add_component(
            "tau",
            Var(
                b.component_pair_set,
                within=Reals,
                initialize=tau_init,
                doc="Binary interaction energy parameters",
                units=pyunits.dimensionless,
            ),
        )

    @staticmethod
    def return_expression(b, pobj, i, j, T):
        if (i, j) in pobj.tau:
            return pobj.tau[i, j]
        elif (j, i) in pobj.tau:
            return pobj.tau[j, i]
        elif i == j:
            return 0
        else:
            raise BurntToast(
                "{} tau rule encountered unexpected index {}. Please contact"
                "the IDAES Developers with this bug.".format(b.name, (i, j))
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Initializer class for implementing Hierarchical initialization routines for
IDAES models with standard forms (e.g. units with 1 control volume)
"""
from pyomo.environ import Block
from pyomo.common.config import ConfigValue, Bool

from idaes.core.initialization.initializer_base import ModularInitializerBase
from idaes.core.util import to_json, from_json, StoreSpec
import idaes.logger as idaeslog
from idaes.core.util.model_statistics import variables_in_activated_constraints_set

__author__ = "Andrew Lee"


class SingleControlVolumeUnitInitializer(ModularInitializerBase):
    """
    This is a general purpose Initializer object for unit models which
    have a single Control Volume named 'control_volume'.

    For details of the initialization routine, please see the documentation.

    """

    CONFIG = ModularInitializerBase.CONFIG()
    CONFIG.declare(
        "always_estimate_states",
        ConfigValue(
            default=False,
            domain=Bool,
            doc="Whether initialization routine should estimate values for "
            "state variables that already have values. Note that if set to True, this will "
            "overwrite any initial guesses provided.",
        ),
    )

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
        copy_inlet_state: bool = False,
    ):
        """
        Common initialization routine for models with one control volume.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.

        Returns:
            Pyomo solver results object
        """
        # The default initialization_routine is sufficient
        return super().initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
            copy_inlet_state=copy_inlet_state,
        )

    def initialize_main_model(self, model: Block, copy_inlet_state: bool = False):
        """
        Initialization routine for models with a single control volume.

        Args:
            model: current model being initialized
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.

        Returns:
            Pyomo solver results object from solve of main model

        """
        # Check to make sure model has something named "control_volume"
        if not hasattr(model, "control_volume"):
            raise TypeError(
                f"Model {model.name} does not appear to be a standard form unit model. "
                f"Please use an Initializer specific to the model being initialized."
            )

        # Get logger
        _log = self.get_logger(model)

        self.initialize_control_volume(model.control_volume, copy_inlet_state)

        # Solve main model
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = self._get_solver().solve(model, tee=slc.tee)

        _log.info_high(f"Control volume initialization {idaeslog.condition(results)}.")

        return results

    def initialize_control_volume(
        self, control_volume: Block, copy_inlet_state: bool = False
    ):
        """
        Initialization routine for control volumes (supports both 0D and 1D control volumes).

        This routine initialized the state and reaction blocks associated with the control volume.

        Args:
            control_volume: control_volume to be initialized
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.

        Returns:
            None

        """
        # Get logger
        _log = self.get_logger(control_volume)

        # Initialize properties
        if hasattr(control_volume, "properties_in"):
            # 0-D control volume
            self._init_props_0D(control_volume, copy_inlet_state)
        else:
            # 1-D control volume
            self._init_props_1D(control_volume)

        _log.info_high("Control volume properties initialization complete")

        # Initialize reactions if they exist
        if hasattr(control_volume, "reactions"):
            self._init_rxns(control_volume)
        _log.info_high("Control volume reactions initialization complete")

    def _init_props_0D(self, control_volume, copy_inlet_state):
        # Initialize inlet properties - inlet state should already be fixed
        prop_init = self.get_submodel_initializer(control_volume.properties_in)

        if prop_init is not None:
            prop_init.initialize(
                model=control_volume.properties_in,
                output_level=self.get_output_level(),
            )

        if not copy_inlet_state:
            # Just in case the user set a different initializer for the outlet
            prop_init = self.get_submodel_initializer(control_volume.properties_out)

            # Estimate missing values for outlet state
            control_volume.estimate_outlet_state(
                always_estimate=self.config.always_estimate_states
            )

            if prop_init is not None:
                prop_init.initialize(
                    model=control_volume.properties_out,
                    output_level=self.get_output_level(),
                )
        else:
            # Map solution from inlet properties to outlet properties
            state = to_json(
                control_volume.properties_in,
                wts=StoreSpec().value(),
                return_dict=True,
            )
            from_json(
                control_volume.properties_out,
                sd=state,
                wts=StoreSpec().value(only_not_fixed=True),
            )

    def _init_props_1D(self, control_volume):
        prop_init = self.get_submodel_initializer(control_volume.properties)

        # Estimate missing values for states
        control_volume.estimate_states(
            always_estimate=self.config.always_estimate_states
        )

        if prop_init is not None:
            prop_init.initialize(
                control_volume.properties,
                output_level=self.get_output_level(),
            )

    def _init_rxns(self, control_volume):
        rxn_init = self.get_submodel_initializer(control_volume.reactions)

        if rxn_init is not None:
            # Reaction blocks depend on vars from other blocks
            # Need to make sure these are all fixed before initializing
            fixed_vars = []
            vset = variables_in_activated_constraints_set(control_volume.reactions)
            for v in vset:
                if v.parent_block().parent_component() is not control_volume.reactions:
                    # Variable external to reactions
                    if not v.fixed:
                        v.fix()
                        fixed_vars.append(v)

            rxn_init.initialize(
                control_volume.reactions,
                output_level=self.get_output_level(),
            )

            # Unfix any vars we fixed previously
            for v in fixed_vars:
                v.unfix()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module tests for the VarLikeExpression component.
"""

import pytest
import re

from pyomo.environ import ConcreteModel, Expression, Var

from idaes.core.base.var_like_expression import VarLikeExpression


@pytest.mark.unit
def test_SimpleVarLikeExpression():
    m = ConcreteModel()

    # Need a Var to use in the Expression to avoid being able to set the value
    # of a float
    m.v = Var(initialize=42)

    m.e = VarLikeExpression(expr=m.v)

    assert m.e.ctype is Expression
    assert not m.e.is_indexed()

    with pytest.raises(
        TypeError,
        match=re.escape(
            "e is an Expression and does not have a value "
            "attribute. Use the 'value()' method instead."
        ),
    ):
        assert m.e.value == 42

    with pytest.raises(
        TypeError,
        match="e is an Expression and does not have a value which can be set.",
    ):
        m.e.set_value(10)

    with pytest.raises(
        TypeError,
        match="e is an Expression and does not have a value which can be set.",
    ):
        m.e.value = 10

    with pytest.raises(
        TypeError,
        match="e is an Expression and can not have bounds. "
        "Use an inequality Constraint instead.",
    ):
        m.e.setub(10)
    with pytest.raises(
        TypeError,
        match="e is an Expression and can not have bounds. "
        "Use an inequality Constraint instead.",
    ):
        m.e.setlb(0)
    with pytest.raises(
        TypeError,
        match="e is an Expression and can not be fixed. "
        "Use an equality Constraint instead.",
    ):
        m.e.fix(8)
    with pytest.raises(TypeError, match="e is an Expression and can not be unfixed."):
        m.e.unfix()

    m.e.set_value(10, force=True)
    assert m.e.expr == 10


@pytest.mark.unit
def test_IndexedVarLikeExpression():
    m = ConcreteModel()

    # Need a Var to use in the Expression to avoid being able to set the value
    # of a float
    m.v = Var(initialize=42)

    m.e = VarLikeExpression([1, 2, 3, 4], expr=m.v)

    assert m.e.ctype is Expression
    assert m.e.is_indexed()

    with pytest.raises(
        TypeError,
        match="e is an Expression and can not have bounds. "
        "Use inequality Constraints instead.",
    ):
        m.e.setub(10)
    with pytest.raises(
        TypeError,
        match="e is an Expression and can not have bounds. "
        "Use inequality Constraints instead.",
    ):
        m.e.setlb(0)
    with pytest.raises(
        TypeError,
        match="e is an Expression and can not be fixed. "
        "Use equality Constraints instead.",
    ):
        m.e.fix(8)
    with pytest.raises(TypeError, match="e is an Expression and can not be unfixed."):
        m.e.unfix()

    for i in m.e:
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and does not have"
                f" a value attribute. Use the 'value()' method instead"
            ),
        ):
            assert m.e[i].value == 42

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and does not "
                f"have a value which can be set."
            ),
        ):
            m.e[i].set_value(10)

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and does not have "
                "a value which can be set."
            ),
        ):
            m.e[i].value = 10

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and can not have "
                "bounds. Use an inequality Constraint instead."
            ),
        ):
            m.e[i].setub(10)
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and can not have "
                "bounds. Use an inequality Constraint instead."
            ),
        ):
            m.e[i].setlb(0)
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"e[{i}] is an Expression and can not be "
                "fixed. Use an equality Constraint instead."
            ),
        ):
            m.e[i].fix(8)
        with pytest.raises(
            TypeError,
            match=re.escape(f"e[{i}] is an Expression and can not be unfixed."),
        ):
            m.e[i].unfix()

        m.e[i].set_value(i, force=True)
        assert m.e[i].expr == i



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""
import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Expression,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.core import LiquidPhase, VaporPhase

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.state_definitions import FcTP
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    IdealBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity

import idaes.models.properties.modular_properties.pure.Perrys as Perrys
import idaes.models.properties.modular_properties.pure.RPP4 as RPP4
import idaes.models.properties.modular_properties.pure.NIST as NIST

from idaes.models.properties.tests.test_harness import PropertyTestHarness


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


config_dict = {
    "components": {
        "benzene": {
            "type": Component,
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": 78.1136e-3,  # [1]
                "pressure_crit": 48.9e5,  # [1]
                "temperature_crit": 562.2,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": 1.0162,  # [2] pg. 2-98
                    "2": 0.2655,
                    "3": 562.16,
                    "4": 0.28212,
                },
                "cp_mol_ig_comp_coeff": {
                    "A": -3.392e1,  # [1]
                    "B": 4.739e-1,
                    "C": -3.017e-4,
                    "D": 7.130e-8,
                },
                "cp_mol_liq_comp_coeff": {
                    "1": 1.29e2,  # [2]
                    "2": -1.7e-1,
                    "3": 6.48e-4,
                    "4": 0,
                    "5": 0,
                },
                "enth_mol_form_liq_comp_ref": 49.0e3,  # [3]
                "enth_mol_form_vap_comp_ref": 82.9e3,  # [3]
                "pressure_sat_comp_coeff": {
                    "A": 4.72583,  # [NIST]
                    "B": 1660.652,
                    "C": -1.461,
                },
            },
        },
        "toluene": {
            "type": Component,
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": 92.1405e-3,  # [1]
                "pressure_crit": 41e5,  # [1]
                "temperature_crit": 591.8,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": 0.8488,  # [2] pg. 2-98
                    "2": 0.26655,
                    "3": 591.8,
                    "4": 0.2878,
                },
                "cp_mol_ig_comp_coeff": {
                    "A": -2.435e1,
                    "B": 5.125e-1,
                    "C": -2.765e-4,
                    "D": 4.911e-8,
                },
                "cp_mol_liq_comp_coeff": {
                    "1": 1.40e2,  # [2]
                    "2": -1.52e-1,
                    "3": 6.95e-4,
                    "4": 0,
                    "5": 0,
                },
                "enth_mol_form_liq_comp_ref": 12.0e3,  # [3]
                "enth_mol_form_vap_comp_ref": 50.1e3,  # [3]
                "pressure_sat_comp_coeff": {
                    "A": 4.07827,  # [NIST]
                    "B": 1343.943,
                    "C": -53.773,
                },
            },
        },
    },
    "phases": {
        "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
    },
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "state_definition": FcTP,
    "state_bounds": {
        "flow_mol_comp": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 450, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew,
}


class TestBTIdeal_FcTP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = GenericParameterBlock
        self.param_args = config_dict
        self.prop_args = {}
        self.has_density_terms = False


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**config_dict)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 2
        for i in model.params.component_list:
            assert i in ["benzene", "toluene"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 4
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "benzene"),
                ("Liq", "toluene"),
                ("Vap", "benzene"),
                ("Vap", "toluene"),
            ]

        assert model.params.config.state_definition == FcTP

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            {
                "flow_mol_comp": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 450, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 2
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["benzene", ("Vap", "Liq")],
            "PE2": ["toluene", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**config_dict)

        model.props = model.params.state_block_class(
            [1], parameters=model.params, defined_state=True
        )

        model.props[1].calculate_scaling_factors()

        # Fix state
        model.props[1].flow_mol_comp.fix(0.5)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Expression)
        assert isinstance(model.props[1].flow_mol_comp, Var)
        for j in model.params.component_list:
            assert value(model.props[1].flow_mol_comp[j]) == 0.5
            assert model.props[1].flow_mol_comp[j].ub == 1000
            assert model.props[1].flow_mol_comp[j].lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325
        assert model.props[1].pressure.ub == 1e6
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 368
        assert model.props[1].temperature.ub == 450
        assert model.props[1].temperature.lb == 273.15

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 2
        for i in model.props[1].mole_frac_comp:
            assert value(model.props[1].mole_frac_comp[i]) == 0.5

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_basic_scaling(self, model):

        assert len(model.props[1].scaling_factor) == 25
        assert model.props[1].scaling_factor[model.props[1].flow_mol] == 1e-2
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_phase["Liq"]] == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_phase["Vap"]] == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_comp["benzene"]]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_comp["toluene"]]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Liq", "benzene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Liq", "toluene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Vap", "benzene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Vap", "toluene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].mole_frac_comp["benzene"]]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[model.props[1].mole_frac_comp["toluene"]]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Liq", "toluene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Vap", "toluene"]
            ]
            == 1000
        )
        assert model.props[1].scaling_factor[model.props[1].pressure] == 1e-5
        assert model.props[1].scaling_factor[model.props[1].temperature] == 1e-2
        assert model.props[1].scaling_factor[model.props[1]._teq["Vap", "Liq"]] == 1e-2
        assert model.props[1].scaling_factor[model.props[1]._t1_Vap_Liq] == 1e-2

        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tbub["Vap", "Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tbub["Vap", "Liq", "toluene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tdew["Vap", "Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tdew["Vap", "Liq", "toluene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].temperature_bubble["Vap", "Liq"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].temperature_dew["Vap", "Liq"]]
            == 1e-2
        )

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 3
        for i in sv:
            assert i in ["flow_mol_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 3
        for i in sv:
            assert i in ["flow_mol_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 3
        for i in sv:
            assert i in ["Molar Flowrate", "Temperature", "Pressure"]

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "benzene"
        ].value == pytest.approx(0.4070, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "benzene"
        ].value == pytest.approx(0.6296, abs=1e-4)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0.4177, abs=1e-4)

        assert value(model.props[1].enth_mol) == pytest.approx(48220.5, 1e-5)

    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""This is an example supercritical pulverized coal (SCPC) power plant steam cycle
model.  This model doesn't represent any specific power plant, but it represents
what could be considered a typical SCPC plant, producing around 620 MW gross.
This model is for demonstration and tutorial purposes only. Before looking at the
model, it may be useful to look at the process flow diagram (PFD).
"""
# Model needs to access private flow terms
# pylint: disable=protected-access

__author__ = "John Eslick, Maojian Wang"

# Import Python libraries
from collections import OrderedDict
import os
import argparse
import logging

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.network import Arc, Port
from pyomo.common.fileutils import this_file_dir

# IDAES Imports
from idaes.core import FlowsheetBlock  # Flowsheet class
from idaes.core.util import model_serializer as ms  # load/save model state
from idaes.core.util.tags import svg_tag, ModelTagGroup  # place numbers/text in an SVG
from idaes.models.properties import iapws95  # steam properties
from idaes.models_extra.power_generation.unit_models.helm import (
    HelmTurbineMultistage,
    HelmMixer,
    HelmIsentropicCompressor,
    HelmTurbineStage,
    HelmNtuCondenser as Condenser,
)
from idaes.models_extra.power_generation.unit_models import FWH0D
from idaes.models.unit_models import (  # basic IDAES unit models, and enum
    MomentumMixingType,  # Enum type for mixer pressure calculation selection
)
from idaes.core.util.initialization import (
    propagate_state as _set_port,
)  # for model initialization
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom

# Callback used to construct heat exchangers with the Underwood approx. for LMTD
from idaes.models.unit_models.heat_exchanger import (
    delta_temperature_underwood_callback,
)
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale


_log = idaeslog.getModelLogger(__name__, logging.INFO)


def create_model():
    """Create the flowsheet and add unit models. Fixing model inputs is done
    in a separate function to try to keep this fairly clean and easy to follow.

    Args:
        None

    Returns:
        (ConcreteModel) Steam cycle model
    """
    ############################################################################
    #  Flowsheet and Properties                                                #
    ############################################################################
    m = pyo.ConcreteModel(name="Steam Cycle Model")
    m.fs = FlowsheetBlock(dynamic=False)  # Add steady state flowsheet

    # A physical property parameter block for IAPWS-95 with pressure and enthalpy
    # (PH) state variables.  Usually pressure and enthalpy state variables are
    # more robust especially when the phases are unknown.
    m.fs.prop_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.MIX
    )

    # A physical property parameter block with temperature, pressure and vapor
    # fraction (TPx) state variables. There are a few instances where the vapor
    # fraction is known and the temperature and pressure state variables are
    # preferable.
    m.fs.prop_water_tpx = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG, state_vars=iapws95.StateVars.TPX
    )
    ############################################################################
    #  Turbine with fill-in reheat constraints                                 #
    ############################################################################
    # The TurbineMultistage class allows creation of the full turbine model by
    # providing several configuration options, including: throttle valves;
    # high-, intermediate-, and low-pressure sections; steam extractions; and
    # pressure driven flow.  See the IDAES documentation for details.
    m.fs.turb = HelmTurbineMultistage(
        property_package=m.fs.prop_water,
        num_parallel_inlet_stages=4,
        num_hp=7,
        num_ip=10,
        num_lp=11,
        hp_split_locations=[4, 7],
        ip_split_locations=[5, 10],
        lp_split_locations=[4, 8, 10, 11],
        hp_disconnect=[7],
        ip_split_num_outlets={10: 3},
    )
    # This model is only the steam cycle, and the reheater is part of the boiler.
    # To fill in the reheater gap, a few constraints for the flow, pressure drop,
    # and outlet temperature are added. A detailed boiler model can be coupled later.
    #
    # hp_split[7] is the splitter directly after the last HP stage.  The splitter
    # outlet "outlet_1" is always taken to be the main steam flow through the turbine.
    # When the turbine model was instantiated the stream from the HP section to the IP
    # section was omitted, so the reheater could be inserted.

    # The flow constraint sets flow from outlet_1 of the splitter equal to
    # flow into the IP turbine.
    @m.fs.turb.Constraint(m.fs.time)
    def constraint_reheat_flow(b, t):
        return b.ip_stages[1].inlet.flow_mol[t] == b.hp_split[7].outlet_1.flow_mol[t]

    # Create a variable for pressure change in the reheater (assuming
    # reheat_delta_p should be negative).
    m.fs.turb.reheat_delta_p = pyo.Var(m.fs.time, initialize=0, units=pyo.units.Pa)

    # Add a constraint to calculate the IP section inlet pressure based on the
    # pressure drop in the reheater and the outlet pressure of the HP section.
    @m.fs.turb.Constraint(m.fs.time)
    def constraint_reheat_press(b, t):
        return (
            b.ip_stages[1].inlet.pressure[t]
            == b.hp_split[7].outlet_1.pressure[t] + b.reheat_delta_p[t]
        )

    # Create a variable for reheat temperature and fix it to the desired reheater
    # outlet temperature
    m.fs.turb.reheat_out_T = pyo.Var(m.fs.time, initialize=866, units=pyo.units.K)

    # Create a constraint for the IP section inlet temperature.
    @m.fs.turb.Constraint(m.fs.time)
    def constraint_reheat_temp(b, t):
        return (
            b.ip_stages[1].control_volume.properties_in[t].temperature
            == b.reheat_out_T[t]
        )

    ############################################################################
    #  Add Condenser/hotwell/condensate pump                                   #
    ############################################################################
    # Add a mixer for all the streams coming into the condenser.  In this case the
    # main steam, and the boiler feed pump turbine outlet go to the condenser
    m.fs.condenser_mix = HelmMixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["main", "bfpt"],
        property_package=m.fs.prop_water,
    )

    # The pressure in the mixer comes from the connection to the condenser.  All
    # the streams coming in and going out of the mixer are equal, but we created
    # the mixer with no calculation for the unit pressure. Here a constraint that
    # specifies that the mixer pressure is equal to the main steam pressure is
    # added.  There is also a constraint that specifies the that BFP turbine outlet
    # pressure is the same as the condenser pressure.  Combined with the stream
    # connections between units, these constraints effectively specify that the
    # mixer inlet and outlet streams all have the same pressure.
    @m.fs.condenser_mix.Constraint(m.fs.time)
    def mixer_pressure_constraint(b, t):
        return b.main_state[t].pressure == b.mixed_state[t].pressure

    # PYLINT-TODO the name "mixer_pressure_constraint" is reused below to define other constraint functions,
    # causing pylint to report function-redefined errors
    # this likely does not actually cause issues at runtime,
    # but it could be worth to check anyway if the pylint errors can be addressed
    # e.g. by giving a unique name to each of the affected functions
    # pylint: disable=function-redefined

    # The condenser model uses the physical property model with TPx state
    # variables, while the rest of the model uses PH state variables. To
    # translate between the two property calculations, an extra port is added to
    # the mixer which contains temperature, pressure, and vapor fraction
    # quantities.
    m.fs.condenser_mix._flow_mol_ref = pyo.Reference(
        m.fs.condenser_mix.mixed_state[:].flow_mol
    )
    m.fs.condenser_mix._temperature_ref = pyo.Reference(
        m.fs.condenser_mix.mixed_state[:].temperature
    )
    m.fs.condenser_mix._pressure_ref = pyo.Reference(
        m.fs.condenser_mix.mixed_state[:].pressure
    )
    m.fs.condenser_mix._vapor_frac_ref = pyo.Reference(
        m.fs.condenser_mix.mixed_state[:].vapor_frac
    )

    m.fs.condenser_mix.outlet_tpx = Port(
        initialize={
            "flow_mol": m.fs.condenser_mix._flow_mol_ref,
            "temperature": m.fs.condenser_mix._temperature_ref,
            "pressure": m.fs.condenser_mix._pressure_ref,
            "vapor_frac": m.fs.condenser_mix._vapor_frac_ref,
        }
    )

    # Add NTU condenser model
    m.fs.condenser = Condenser(
        dynamic=False,
        shell={"has_pressure_change": False, "property_package": m.fs.prop_water},
        tube={"has_pressure_change": False, "property_package": m.fs.prop_water},
    )

    # Add the condenser hotwell.  In steady state a mixer will work.  This is
    # where makeup water is added if needed.
    m.fs.hotwell = HelmMixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["condensate", "makeup"],
        property_package=m.fs.prop_water,
    )

    # The hotwell is assumed to be at the same pressure as the condenser.
    @m.fs.hotwell.Constraint(m.fs.time)
    def mixer_pressure_constraint(b, t):
        return b.condensate_state[t].pressure == b.mixed_state[t].pressure

    # Condensate pump (Use compressor model, since it is more robust if vapor form)
    m.fs.cond_pump = HelmIsentropicCompressor(property_package=m.fs.prop_water)
    ############################################################################
    #  Add low pressure feedwater heaters                                      #
    ############################################################################
    # All the feedwater heater sections will be set to use the Underwood
    # approximation for LMTD, so create the fwh_config dict to make the config
    # slightly cleaner
    fwh_config = {"delta_temperature_callback": delta_temperature_underwood_callback}

    # The feedwater heater model allows feedwater heaters with a desuperheat,
    # condensing, and subcooling section to be added an a reasonably simple way.
    # See the IDAES documentation for more information of configuring feedwater
    # heaters
    m.fs.fwh1 = FWH0D(
        has_desuperheat=False,
        has_drain_cooling=False,
        has_drain_mixer=True,
        property_package=m.fs.prop_water,
        condense=fwh_config,
    )
    # pump for fwh1 condensate, to pump it ahead and mix with feedwater
    m.fs.fwh1_pump = HelmIsentropicCompressor(property_package=m.fs.prop_water)
    # Mix the FWH1 drain back into the feedwater
    m.fs.fwh1_return = HelmMixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["feedwater", "fwh1_drain"],
        property_package=m.fs.prop_water,
    )

    # Set the mixer pressure to the feedwater pressure
    @m.fs.fwh1_return.Constraint(m.fs.time)
    def mixer_pressure_constraint(b, t):
        return b.feedwater_state[t].pressure == b.mixed_state[t].pressure

    # Add the rest of the low pressure feedwater heaters
    m.fs.fwh2 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    m.fs.fwh3 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    m.fs.fwh4 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=False,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    ############################################################################
    #  Add deaerator and boiler feed pump (BFP)                                #
    ############################################################################
    # The deaerator is basically an open tank with multiple inlets.  For steady-
    # state, a mixer model is sufficient.
    m.fs.fwh5_da = HelmMixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["steam", "drain", "feedwater"],
        property_package=m.fs.prop_water,
    )

    @m.fs.fwh5_da.Constraint(m.fs.time)
    def mixer_pressure_constraint(b, t):
        # Not sure about deaerator pressure, so assume same as feedwater inlet
        return b.feedwater_state[t].pressure == b.mixed_state[t].pressure

    # Add the boiler feed pump and boiler feed pump turbine
    m.fs.bfp = HelmIsentropicCompressor(property_package=m.fs.prop_water)

    m.fs.bfpt = HelmTurbineStage(property_package=m.fs.prop_water)

    # The boiler feed pump outlet pressure is the same as the condenser
    @m.fs.Constraint(m.fs.time)
    def constraint_out_pressure(b, t):
        return (
            b.bfpt.control_volume.properties_out[t].pressure
            == b.condenser.shell.properties_out[t].pressure
        )

    # Instead of specifying a fixed efficiency, specify that the steam is just
    # starting to condense at the outlet of the boiler feed pump turbine.  This
    # ensures approximately the right behavior in the turbine.  With a fixed
    # efficiency, depending on the conditions you can get odd things like steam
    # fully condensing in the turbine.
    @m.fs.Constraint(m.fs.time)
    def constraint_out_enthalpy(b, t):
        return (
            b.bfpt.control_volume.properties_out[t].enth_mol
            == b.bfpt.control_volume.properties_out[t].enth_mol_sat_phase["Vap"]
            - 200 * pyo.units.J / pyo.units.mol
        )

    # The boiler feed pump power is the same as the power generated by the
    # boiler feed pump turbine. This constraint determines the steam flow to the
    # BFP turbine. The turbine work is negative for power out, while pump work
    # is positive for power in.
    @m.fs.Constraint(m.fs.time)
    def constraint_bfp_power(b, t):
        return 0 == b.bfp.control_volume.work[t] + b.bfpt.control_volume.work[t]

    ############################################################################
    #  Add high pressure feedwater heaters                                     #
    ############################################################################
    m.fs.fwh6 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    m.fs.fwh7 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    m.fs.fwh8 = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=False,
        property_package=m.fs.prop_water,
        desuperheat=fwh_config,
        cooling=fwh_config,
        condense=fwh_config,
    )
    ############################################################################
    #  Additional Constraints/Expressions                                      #
    ############################################################################

    # Add a few constraints to allow a for complete plant results despite the
    # lack of a detailed boiler model.

    # Boiler pressure drop
    m.fs.boiler_pressure_drop_fraction = pyo.Var(
        m.fs.time,
        initialize=0.01,
        doc="Fraction of pressure lost from boiler feed pump and turbine inlet",
    )

    @m.fs.Constraint(m.fs.time)
    def boiler_pressure_drop(b, t):
        return (
            m.fs.bfp.control_volume.properties_out[t].pressure
            * (1 - b.boiler_pressure_drop_fraction[t])
            == m.fs.turb.inlet_split.mixed_state[t].pressure
        )

    # Again, since the boiler is missing, set the flow of steam into the turbine
    # equal to the flow of feedwater out of the last feedwater heater.
    @m.fs.Constraint(m.fs.time)
    def close_flow(b, t):
        return (
            m.fs.bfp.control_volume.properties_out[t].flow_mol
            == m.fs.turb.inlet_split.mixed_state[t].flow_mol
        )

    # Calculate the amount of heat that is added in the boiler, including the
    # reheater.
    @m.fs.Expression(m.fs.time)
    def boiler_heat(b, t):
        return (
            b.turb.inlet_split.mixed_state[t].enth_mol
            * b.turb.inlet_split.mixed_state[t].flow_mol
            - b.fwh8.desuperheat.cold_side.properties_out[t].enth_mol
            * b.fwh8.desuperheat.cold_side.properties_out[t].flow_mol
            + b.turb.ip_stages[1].control_volume.properties_in[t].enth_mol
            * b.turb.ip_stages[1].control_volume.properties_in[t].flow_mol
            - b.turb.hp_split[7].outlet_1.enth_mol[t]
            * b.turb.hp_split[7].outlet_1.flow_mol[t]
        )

    # Calculate the efficiency of the steam cycle.  This doesn't account for
    # heat loss in the boiler, so actual plant efficiency would be lower.
    @m.fs.Expression(m.fs.time)
    def steam_cycle_eff(b, t):
        return -100 * b.turb.power[t] / b.boiler_heat[t]

    ############################################################################
    ##  Create the stream Arcs                                                ##
    ############################################################################

    ############################################################################
    #  Connect turbine and condenser units                                     #
    ############################################################################
    m.fs.EXHST_MAIN = Arc(
        source=m.fs.turb.outlet_stage.outlet, destination=m.fs.condenser_mix.main
    )

    m.fs.condenser_mix_to_condenser = Arc(
        source=m.fs.condenser_mix.outlet, destination=m.fs.condenser.hot_side_inlet
    )

    m.fs.COND_01 = Arc(
        source=m.fs.condenser.hot_side_outlet, destination=m.fs.hotwell.condensate
    )

    m.fs.COND_02 = Arc(source=m.fs.hotwell.outlet, destination=m.fs.cond_pump.inlet)
    ############################################################################
    #  Low pressure FWHs                                                       #
    ############################################################################
    m.fs.EXTR_LP11 = Arc(
        source=m.fs.turb.lp_split[11].outlet_2, destination=m.fs.fwh1.drain_mix.steam
    )
    m.fs.COND_03 = Arc(
        source=m.fs.cond_pump.outlet, destination=m.fs.fwh1.condense.cold_side_inlet
    )
    m.fs.FWH1_DRN1 = Arc(
        source=m.fs.fwh1.condense.hot_side_outlet, destination=m.fs.fwh1_pump.inlet
    )
    m.fs.FWH1_DRN2 = Arc(
        source=m.fs.fwh1_pump.outlet, destination=m.fs.fwh1_return.fwh1_drain
    )
    m.fs.FW01A = Arc(
        source=m.fs.fwh1.condense.cold_side_outlet,
        destination=m.fs.fwh1_return.feedwater,
    )
    # fwh2
    m.fs.FW01B = Arc(
        source=m.fs.fwh1_return.outlet, destination=m.fs.fwh2.cooling.cold_side_inlet
    )
    m.fs.FWH2_DRN = Arc(
        source=m.fs.fwh2.cooling.hot_side_outlet, destination=m.fs.fwh1.drain_mix.drain
    )
    m.fs.EXTR_LP10 = Arc(
        source=m.fs.turb.lp_split[10].outlet_2,
        destination=m.fs.fwh2.desuperheat.hot_side_inlet,
    )
    # fwh3
    m.fs.FW02 = Arc(
        source=m.fs.fwh2.desuperheat.cold_side_outlet,
        destination=m.fs.fwh3.cooling.cold_side_inlet,
    )
    m.fs.FWH3_DRN = Arc(
        source=m.fs.fwh3.cooling.hot_side_outlet, destination=m.fs.fwh2.drain_mix.drain
    )
    m.fs.EXTR_LP8 = Arc(
        source=m.fs.turb.lp_split[8].outlet_2,
        destination=m.fs.fwh3.desuperheat.hot_side_inlet,
    )
    # fwh4
    m.fs.FW03 = Arc(
        source=m.fs.fwh3.desuperheat.cold_side_outlet,
        destination=m.fs.fwh4.cooling.cold_side_inlet,
    )
    m.fs.FWH4_DRN = Arc(
        source=m.fs.fwh4.cooling.hot_side_outlet, destination=m.fs.fwh3.drain_mix.drain
    )
    m.fs.EXTR_LP4 = Arc(
        source=m.fs.turb.lp_split[4].outlet_2,
        destination=m.fs.fwh4.desuperheat.hot_side_inlet,
    )
    ############################################################################
    #  FWH5 (Deaerator) and boiler feed pump (BFP)                             #
    ############################################################################
    m.fs.FW04 = Arc(
        source=m.fs.fwh4.desuperheat.cold_side_outlet,
        destination=m.fs.fwh5_da.feedwater,
    )
    m.fs.EXTR_IP10 = Arc(
        source=m.fs.turb.ip_split[10].outlet_2, destination=m.fs.fwh5_da.steam
    )
    m.fs.FW05A = Arc(source=m.fs.fwh5_da.outlet, destination=m.fs.bfp.inlet)
    m.fs.EXTR_BFPT_A = Arc(
        source=m.fs.turb.ip_split[10].outlet_3, destination=m.fs.bfpt.inlet
    )
    m.fs.EXHST_BFPT = Arc(source=m.fs.bfpt.outlet, destination=m.fs.condenser_mix.bfpt)
    ############################################################################
    #  High-pressure feedwater heaters                                         #
    ############################################################################
    # fwh6
    m.fs.FW05B = Arc(
        source=m.fs.bfp.outlet, destination=m.fs.fwh6.cooling.cold_side_inlet
    )
    m.fs.FWH6_DRN = Arc(
        source=m.fs.fwh6.cooling.hot_side_outlet, destination=m.fs.fwh5_da.drain
    )
    m.fs.EXTR_IP5 = Arc(
        source=m.fs.turb.ip_split[5].outlet_2,
        destination=m.fs.fwh6.desuperheat.hot_side_inlet,
    )
    # fwh7
    m.fs.FW06 = Arc(
        source=m.fs.fwh6.desuperheat.cold_side_outlet,
        destination=m.fs.fwh7.cooling.cold_side_inlet,
    )
    m.fs.FWH7_DRN = Arc(
        source=m.fs.fwh7.cooling.hot_side_outlet, destination=m.fs.fwh6.drain_mix.drain
    )
    m.fs.EXTR_HP7 = Arc(
        source=m.fs.turb.hp_split[7].outlet_2,
        destination=m.fs.fwh7.desuperheat.hot_side_inlet,
    )
    # fwh8
    m.fs.FW07 = Arc(
        source=m.fs.fwh7.desuperheat.cold_side_outlet,
        destination=m.fs.fwh8.cooling.cold_side_inlet,
    )
    m.fs.FWH8_DRN = Arc(
        source=m.fs.fwh8.cooling.hot_side_outlet, destination=m.fs.fwh7.drain_mix.drain
    )
    m.fs.EXTR_HP4 = Arc(
        source=m.fs.turb.hp_split[4].outlet_2,
        destination=m.fs.fwh8.desuperheat.hot_side_inlet,
    )

    ############################################################################
    # Turn the Arcs into constraints and return the model                      #
    ############################################################################
    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)
    return m


def _stream_dict(m):
    """Adds _streams to m, which contains a dictionary of streams for display

    Args:
        m (ConcreteModel): A Pyomo model from create_model()

    Returns:
        None
    """

    # Put all the top-level streams in the stream dictionary
    m._streams = dict(
        [(c.getname(), c) for c in m.fs.component_objects(Arc, descend_into=False)]
    )

    # There are some additional streams we are interested in that are either
    # inlets or outlets, where there is no arc or arcs buried in unit models.
    # Next those are added specifically.
    m._streams.update(
        {
            "STEAM_MAIN": m.fs.turb.inlet_split.mixed_state,
            "THRTL1": m.fs.turb.inlet_stage[1].control_volume.properties_in,
            "THRTL2": m.fs.turb.inlet_stage[2].control_volume.properties_in,
            "THRTL3": m.fs.turb.inlet_stage[3].control_volume.properties_in,
            "THRTL4": m.fs.turb.inlet_stage[4].control_volume.properties_in,
            "RHT_COLD": m.fs.turb.hp_split[7].outlet_1_state,
            "RHT_HOT": m.fs.turb.ip_stages[1].control_volume.properties_in,
            "STEAM_LP": m.fs.turb.ip_split[10].outlet_1_state,
            "CW01": m.fs.condenser.tube.properties_in,
            "CW02": m.fs.condenser.tube.properties_out,
            "MAKEUP_01": m.fs.hotwell.makeup_state,
            "FW08": m.fs.fwh8.desuperheat.cold_side.properties_out,
        }
    )

    # Alphabetize to find streams easier in the tabular stream table view.
    m._streams = OrderedDict(sorted(m._streams.items()))


def set_model_input(m):
    """Fix some variables and set values. Generally get the model ready to run
    in simulation mode (0 degrees of freedom).

    Args:
        m (ConcreteModel): A Pyomo model from create_model()

    Returns:
        None
    """
    ############################################################################
    #  Turbine input                                                           #
    ############################################################################
    main_steam_pressure = 2.423e7
    m.fs.turb.turbine_inlet_cf_fix(4.6e-4)
    m.fs.turb.turbine_outlet_cf_fix(0.32)
    # Set the turbine steam inlet conditions and flow guess for init
    m.fs.turb.inlet_split.inlet.enth_mol.fix(62710)
    m.fs.turb.inlet_split.inlet.pressure.fix(main_steam_pressure)
    m.fs.boiler_pressure_drop_fraction.fix(0.1)
    m.fs.turb.inlet_split.inlet.flow_mol[:].value = 21000
    m.fs.turb.inlet_split.inlet.flow_mol.unfix()  # Pressure-driven
    m.fs.turb.inlet_mix.use_equal_pressure_constraint()  # Pressure-driven mix
    # Set throttle valve
    m.fs.turb.throttle_cv_fix(0.0009)
    m.fs.turb.throttle_valve[1].valve_opening.fix(0.9)
    m.fs.turb.throttle_valve[2].valve_opening.fix(0.9)
    m.fs.turb.throttle_valve[3].valve_opening.fix(0.9)
    m.fs.turb.throttle_valve[4].valve_opening.fix(0.9)
    # Set the efficiency and pressure ratios of stages other than inlet and outlet
    for s in m.fs.turb.hp_stages.values():
        s.ratioP.fix(0.80)
        s.efficiency_isentropic.fix(0.9)
    for s in m.fs.turb.ip_stages.values():
        s.ratioP.fix(0.78)
        s.efficiency_isentropic.fix(0.9)
    for s in m.fs.turb.lp_stages.values():
        s.ratioP[:].fix(0.76)
        s.efficiency_isentropic[:].fix(0.9)
    ############################################################################
    #  Reheater                                                                #
    ############################################################################
    m.fs.turb.reheat_delta_p.fix(0)
    m.fs.turb.reheat_out_T.fix(866)
    ############################################################################
    #  Condenser section inputs                                                #
    ############################################################################
    m.fs.condenser.cold_side_inlet.flow_mol.fix(2500000)
    m.fs.condenser.cold_side_inlet.enth_mol.fix(1700)
    m.fs.condenser.cold_side_inlet.pressure.fix(500000)
    m.fs.condenser.area.fix(13000)
    m.fs.condenser.overall_heat_transfer_coefficient.fix(15000)

    m.fs.hotwell.makeup.flow_mol[:].value = 1  # don't fix is calculated
    m.fs.hotwell.makeup.enth_mol.fix(2500)
    m.fs.hotwell.makeup.pressure.fix(101325)
    m.fs.cond_pump.efficiency_isentropic.fix(0.80)
    m.fs.cond_pump.deltaP.fix(1e6)
    ############################################################################
    #  Low-pressure FWH section inputs                                         #
    ############################################################################
    # fwh1
    # Heat transfer coefficient correlation constraints can be added to the
    # feedwater heaters, but to keep this example simple, they are fixed
    # constant values here.
    m.fs.fwh1.condense.area.fix(400)
    m.fs.fwh1.condense.overall_heat_transfer_coefficient.fix(2000)
    # fwh1 pump
    m.fs.fwh1_pump.efficiency_isentropic.fix(0.80)
    m.fs.fwh1_pump.deltaP.fix(1.2e6)  # need pressure higher than feedwater
    # fwh2
    m.fs.fwh2.condense.area.fix(150)
    m.fs.fwh2.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh2.desuperheat.area.fix(75)
    m.fs.fwh2.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh2.cooling.area.fix(75)
    m.fs.fwh2.cooling.overall_heat_transfer_coefficient.fix(300)
    # fwh3
    m.fs.fwh3.condense.area.fix(100)
    m.fs.fwh3.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh3.desuperheat.area.fix(50)
    m.fs.fwh3.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh3.cooling.area.fix(50)
    m.fs.fwh3.cooling.overall_heat_transfer_coefficient.fix(300)
    # fwh4
    m.fs.fwh4.condense.area.fix(100)
    m.fs.fwh4.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh4.desuperheat.area.fix(50)
    m.fs.fwh4.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh4.cooling.area.fix(50)
    m.fs.fwh4.cooling.overall_heat_transfer_coefficient.fix(300)
    ############################################################################
    #  Deaerator and boiler feed pump (BFP) input                              #
    ############################################################################
    m.fs.bfp.efficiency_isentropic.fix(0.80)
    m.fs.bfp.outlet.pressure[:].value = main_steam_pressure * 1.1  # guess
    m.fs.bfpt.efficiency_isentropic.value = 0.80  # don't fix, just initial guess
    ############################################################################
    #  High-pressure feedwater heater                                          #
    ############################################################################
    # fwh6
    m.fs.fwh6.condense.area.fix(300)
    m.fs.fwh6.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh6.desuperheat.area.fix(150)
    m.fs.fwh6.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh6.cooling.area.fix(150)
    m.fs.fwh6.cooling.overall_heat_transfer_coefficient.fix(300)
    # fwh7
    m.fs.fwh7.condense.area.fix(200)
    m.fs.fwh7.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh7.desuperheat.area.fix(100)
    m.fs.fwh7.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh7.cooling.area.fix(100)
    m.fs.fwh7.cooling.overall_heat_transfer_coefficient.fix(300)
    # fwh8
    m.fs.fwh8.condense.area.fix(200)
    m.fs.fwh8.condense.overall_heat_transfer_coefficient.fix(2000)
    m.fs.fwh8.desuperheat.area.fix(100)
    m.fs.fwh8.desuperheat.overall_heat_transfer_coefficient.fix(600)
    m.fs.fwh8.cooling.area.fix(100)
    m.fs.fwh8.cooling.overall_heat_transfer_coefficient.fix(300)
    # Now all the model input has been specified.


def initialize(m, fileinput=None, outlvl=idaeslog.NOTSET):
    """Initialize a mode from create_model(), set model inputs before
    initializing.

    Args:
        m (ConcreteModel): A Pyomo model from create_model()
        fileinput (str|None): File to load initialized model state from. If a
            file is supplied skip initialization routine. If None, initialize.

    Returns:
        solver: A Pyomo solver object, that can be used to solve the model.

    """
    init_log = idaeslog.getInitLogger(m.name, outlvl, tag="flowsheet")
    solve_log = idaeslog.getSolveLogger(m.name, outlvl, tag="flowsheet")

    # set scaling factors

    iscale.set_scaling_factor(m.fs.condenser.hot_side.heat, 1e-9)
    iscale.set_scaling_factor(m.fs.condenser.cold_side.heat, 1e-9)

    iscale.set_scaling_factor(m.fs.fwh1.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh1.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh2.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh2.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh3.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh3.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh4.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh4.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh6.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh6.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh7.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh7.condense.cold_side.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.fwh8.condense.hot_side.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.fwh8.condense.cold_side.heat, 1e-7)

    iscale.calculate_scaling_factors(m)

    solver = get_solver()

    if fileinput is not None:
        init_log.info("Loading initial values from file: {}".format(fileinput))
        ms.from_json(m, fname=fileinput)
        return solver

    init_log.info("Starting initialization")
    ############################################################################
    #  Initialize turbine                                                      #
    ############################################################################
    # Extraction rates are calculated from the feedwater heater models, so to
    # initialize the turbine fix some initial guesses. They get unfixed after
    # solving the turbine
    m.fs.turb.outlet_stage.control_volume.properties_out[:].pressure.fix(3500)
    m.fs.turb.lp_split[11].split_fraction[:, "outlet_2"].fix(0.04403)
    m.fs.turb.lp_split[10].split_fraction[:, "outlet_2"].fix(0.04025)
    m.fs.turb.lp_split[8].split_fraction[:, "outlet_2"].fix(0.04362)
    m.fs.turb.lp_split[4].split_fraction[:, "outlet_2"].fix(0.08025)
    m.fs.turb.ip_split[10].split_fraction[:, "outlet_2"].fix(0.045)
    m.fs.turb.ip_split[10].split_fraction[:, "outlet_3"].fix(0.04)
    m.fs.turb.ip_split[5].split_fraction[:, "outlet_2"].fix(0.05557)
    m.fs.turb.hp_split[7].split_fraction[:, "outlet_2"].fix(0.09741)
    m.fs.turb.hp_split[4].split_fraction[:, "outlet_2"].fix(0.0740)
    # Put in a rough initial guess for the IP section inlet, since it is
    # disconnected from the HP section for the reheater.
    ip1_pin = 5.35e6
    ip1_hin = pyo.value(iapws95.htpx(T=866 * pyo.units.K, P=ip1_pin * pyo.units.Pa))
    ip1_fin = pyo.value(m.fs.turb.inlet_split.inlet.flow_mol[0])
    m.fs.turb.ip_stages[1].inlet.enth_mol[:].value = ip1_hin
    m.fs.turb.ip_stages[1].inlet.flow_mol[:].value = ip1_fin
    m.fs.turb.ip_stages[1].inlet.pressure[:].value = ip1_pin
    # initialize turbine
    assert degrees_of_freedom(m.fs.turb) == 0
    m.fs.turb.initialize(outlvl=outlvl, optarg=solver.options)
    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
        res = solver.solve(m.fs.turb, tee=slc.tee)
    init_log.info("Full turbine solve complete: {}".format(idaeslog.condition(res)))
    # The turbine outlet pressure is determined by the condenser once hooked up
    m.fs.turb.outlet_stage.control_volume.properties_out[:].pressure.unfix()
    # Extraction rates are calculated once the feedwater heater models are
    # added so unfix the splits.
    m.fs.turb.lp_split[11].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.lp_split[10].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.lp_split[8].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.lp_split[4].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.ip_split[10].split_fraction[:, "outlet_3"].unfix()
    m.fs.turb.ip_split[5].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.hp_split[7].split_fraction[:, "outlet_2"].unfix()
    m.fs.turb.hp_split[4].split_fraction[:, "outlet_2"].unfix()
    # Initialize the boiler feed pump turbine.
    _set_port(m.fs.bfpt.inlet, m.fs.turb.ip_split[10].outlet_3)
    m.fs.bfpt.control_volume.properties_out[:].pressure.fix(10000)
    m.fs.bfpt.efficiency_isentropic.fix()
    m.fs.bfpt.initialize(outlvl=idaeslog.DEBUG, optarg=solver.options)
    m.fs.bfpt.control_volume.properties_out[:].pressure.unfix()
    m.fs.bfpt.efficiency_isentropic.unfix()
    ############################################################################
    #  Condenser section                                                       #
    ############################################################################
    # initialize condenser mixer
    _set_port(m.fs.condenser_mix.main, m.fs.turb.outlet_stage.outlet)
    _set_port(m.fs.condenser_mix.bfpt, m.fs.bfpt.outlet)
    m.fs.condenser_mix.initialize(outlvl=outlvl, optarg=solver.options)
    # initialize condenser hx

    _set_port(m.fs.condenser.hot_side_inlet, m.fs.condenser_mix.outlet)
    _set_port(m.fs.condenser.cold_side_outlet, m.fs.condenser.cold_side_inlet)
    m.fs.condenser.initialize(outlvl=outlvl, optarg=solver.options)

    # initialize hotwell
    _set_port(m.fs.hotwell.condensate, m.fs.condenser.hot_side_outlet)

    m.fs.hotwell.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.hotwell.condensate.unfix()
    # initialize condensate pump
    _set_port(m.fs.cond_pump.inlet, m.fs.hotwell.outlet)
    m.fs.cond_pump.initialize(outlvl=outlvl, optarg=solver.options)
    ############################################################################
    #  Low-pressure FWH section                                                #
    ############################################################################
    # fwh1
    m.fs.fwh1.drain_mix.drain.flow_mol[:] = 1000
    m.fs.fwh1.drain_mix.drain.pressure[:] = 1e5
    m.fs.fwh1.drain_mix.drain.enth_mol[:] = 6117
    _set_port(m.fs.fwh1.condense.cold_side_inlet, m.fs.cond_pump.outlet)
    _set_port(m.fs.fwh1.drain_mix.steam, m.fs.turb.lp_split[11].outlet_2)
    m.fs.fwh1.initialize(outlvl=outlvl, optarg=solver.options)
    # initialize fwh1 drain pump
    _set_port(m.fs.fwh1_pump.inlet, m.fs.fwh1.condense.hot_side_outlet)
    m.fs.fwh1_pump.initialize(outlvl=5, optarg=solver.options)
    # initialize mixer to add fwh1 drain to feedwater
    _set_port(m.fs.fwh1_return.feedwater, m.fs.fwh1.condense.cold_side_outlet)
    _set_port(m.fs.fwh1_return.fwh1_drain, m.fs.fwh1.condense.hot_side_outlet)
    m.fs.fwh1_return.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh1_return.feedwater.unfix()
    m.fs.fwh1_return.fwh1_drain.unfix()
    # fwh2
    m.fs.fwh2.drain_mix.drain.flow_mol[:] = 100
    m.fs.fwh2.drain_mix.drain.pressure[:] = 1.5e5
    m.fs.fwh2.drain_mix.drain.enth_mol[:] = 7000
    _set_port(m.fs.fwh2.cooling.cold_side_inlet, m.fs.fwh1_return.outlet)
    _set_port(m.fs.fwh2.desuperheat.hot_side_inlet, m.fs.turb.lp_split[10].outlet_2)
    m.fs.fwh2.initialize(outlvl=outlvl, optarg=solver.options)
    # fwh3
    m.fs.fwh3.drain_mix.drain.flow_mol[:] = 100
    m.fs.fwh3.drain_mix.drain.pressure[:] = 2.5e5
    m.fs.fwh3.drain_mix.drain.enth_mol[:] = 8000
    _set_port(m.fs.fwh3.cooling.cold_side_inlet, m.fs.fwh2.desuperheat.cold_side_outlet)
    _set_port(m.fs.fwh3.desuperheat.hot_side_inlet, m.fs.turb.lp_split[8].outlet_2)
    m.fs.fwh3.initialize(outlvl=outlvl, optarg=solver.options)
    # fwh4
    _set_port(m.fs.fwh4.cooling.cold_side_inlet, m.fs.fwh3.desuperheat.cold_side_outlet)
    _set_port(m.fs.fwh4.desuperheat.hot_side_inlet, m.fs.turb.lp_split[4].outlet_2)
    m.fs.fwh4.initialize(outlvl=outlvl, optarg=solver.options)
    ############################################################################
    #  boiler feed pump and deaerator                                          #
    ############################################################################
    _set_port(m.fs.fwh5_da.feedwater, m.fs.fwh4.desuperheat.cold_side_outlet)
    _set_port(m.fs.fwh5_da.steam, m.fs.turb.ip_split[10].outlet_2)
    m.fs.fwh5_da.drain.flow_mol[:] = 2000
    m.fs.fwh5_da.drain.pressure[:] = 3e6
    m.fs.fwh5_da.drain.enth_mol[:] = 9000
    m.fs.fwh5_da.initialize(outlvl=outlvl, optarg=solver.options)
    _set_port(m.fs.bfp.inlet, m.fs.fwh5_da.outlet)
    m.fs.bfp.control_volume.properties_out[:].pressure.fix()
    m.fs.bfp.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.bfp.control_volume.properties_out[:].pressure.unfix()
    ############################################################################
    #  High-pressure feedwater heaters                                         #
    ############################################################################
    # fwh6
    m.fs.fwh6.drain_mix.drain.flow_mol[:] = 1000
    m.fs.fwh6.drain_mix.drain.pressure[:] = 1e7
    m.fs.fwh6.drain_mix.drain.enth_mol[:] = 9500
    _set_port(m.fs.fwh6.cooling.cold_side_inlet, m.fs.bfp.outlet)
    _set_port(m.fs.fwh6.desuperheat.hot_side_inlet, m.fs.turb.ip_split[5].outlet_2)
    m.fs.fwh6.initialize(outlvl=outlvl, optarg=solver.options)
    # fwh7
    m.fs.fwh7.drain_mix.drain.flow_mol[:] = 2000
    m.fs.fwh7.drain_mix.drain.pressure[:] = 1e7
    m.fs.fwh7.drain_mix.drain.enth_mol[:] = 9500
    _set_port(m.fs.fwh7.cooling.cold_side_inlet, m.fs.fwh6.desuperheat.cold_side_outlet)
    _set_port(m.fs.fwh7.desuperheat.hot_side_inlet, m.fs.turb.hp_split[7].outlet_2)
    m.fs.fwh7.initialize(outlvl=outlvl, optarg=solver.options)
    # fwh8
    _set_port(m.fs.fwh8.cooling.cold_side_inlet, m.fs.fwh7.desuperheat.cold_side_outlet)
    _set_port(m.fs.fwh8.desuperheat.hot_side_inlet, m.fs.turb.hp_split[4].outlet_2)
    m.fs.fwh8.initialize(outlvl=outlvl, optarg=solver.options)
    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
        res = solver.solve(m, tee=slc.tee)
    init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

    return solver


def pfd_result(m, df, svg):
    """Insert model results into the PFD and return a new SVG string, which
    can be displayed, further edited, or saved to a file.

    Args:
        m (ConcreteModel): A steam cycle model
        df (Pandas DataFrame): Stream table
        svg (FILE*, str, bytes): Original svg as either a file-like object,
            a string, or a byte array.

    Returns:
        (str): SVG content.
    """
    tags = {}  # dict of tags and data to insert into SVG
    for i in df.index:  # Create entries for streams
        tags[i + "_F"] = df.loc[i, "Molar Flow"]
        tags[i + "_T"] = df.loc[i, "T"]
        tags[i + "_P"] = df.loc[i, "P"]
        tags[i + "_X"] = df.loc[i, "Vapor Fraction"]
    # Add some additional quntities from the model to report
    tags["gross_power"] = -pyo.value(m.fs.turb.power[0])
    tags["gross_power_mw"] = -pyo.value(m.fs.turb.power[0]) * 1e-6
    tags["steam_mass_flow"] = df.loc["STEAM_MAIN", "Mass Flow"]
    tags["sc_eff"] = pyo.value(m.fs.steam_cycle_eff[0])
    tags["boiler_heat"] = pyo.value(m.fs.boiler_heat[0]) * 1e-6
    tags["steam_pressure"] = df.loc["STEAM_MAIN", "P"] / 1000.0
    tags["cond_pressure"] = df.loc["EXHST_MAIN", "P"] / 1000.0
    tags["bfp_power"] = pyo.value(m.fs.bfp.work_mechanical[0])
    tags["bfp_eff"] = pyo.value(m.fs.bfp.efficiency_isentropic[0]) * 100
    tags["bfpt_power"] = pyo.value(m.fs.bfpt.work_mechanical[0])
    tags["bfpt_eff"] = pyo.value(m.fs.bfpt.efficiency_isentropic[0]) * 100

    tag_group = ModelTagGroup()
    for t, v in tags.items():
        tag_group.add(t, v, format_string="{:.3f}")
    if svg is None:
        fname = os.path.join(this_file_dir(), "supercritical_steam_cycle.svg")
        with open(fname, "r") as f:
            svg = f.read()
    s = svg_tag(tag_group=tag_group, svg=svg)
    return s


def main(initialize_from_file=None, store_initialization=None):
    """Create and initialize a model and solver

    Args:
        None

    Returns:
        A tuple of a model and solver
    """
    m = create_model()
    _stream_dict(m)
    set_model_input(m)
    if initialize_from_file is None:
        solver = initialize(m)
    else:
        solver = initialize(m, fileinput=initialize_from_file)
    solver.solve(m, tee=True)
    if store_initialization is not None:
        ms.to_json(m, fname=store_initialization)
    return m, solver


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--initialize_from_file",
        help="File from which to load initialized values. If specified, the "
        "initialization procedure will be skipped.",
        default=None,
    )
    parser.add_argument(
        "--store_initialization",
        help="If specified, store" " the initialized model values, to reload later.",
        default=None,
    )
    args = parser.parse_args()
    m, solver = main(
        initialize_from_file=args.initialize_from_file,
        store_initialization=args.store_initialization,
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Predefined expressions for Helmholtz EoS functions
"""

__author__ = "John Eslick"

import pyomo.environ as pyo


def sat_delta_type01(model, name, parameters):
    """Type01 expression for the approximate saturated reduced density

    Args:
        model (Block): Pyomo model
        parameters (dict): Main parameters dictionary

    Returns:
        dict: Expressions for approximate saturated reduced density
    """
    c = parameters["aux"][name]["c"]
    n = parameters["aux"][name]["n"]
    t = parameters["aux"][name]["t"]
    return c + sum(n[i] * (1 - 1 / model.tau) ** t[i] for i in n)


def sat_delta_type02(model, name, parameters):
    """Type01 expression for the approximate saturated reduced density

    Args:
        model (Block): Pyomo model
        parameters (dict): Main parameters dictionary

    Returns:
        dict: Expressions for approximate saturated reduced density
    """
    c = parameters["aux"][name]["c"]
    n = parameters["aux"][name]["n"]
    t = parameters["aux"][name]["t"]
    return c * pyo.exp(sum(n[i] * (1 - 1 / model.tau) ** t[i] for i in n))
